#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer

from argilla.cli.rich import echo_in_panel
from argilla.client.login import ArgillaCredentials
from argilla.client.singleton import init


def init_callback() -> None:
    if not ArgillaCredentials.exists():
        echo_in_panel(
            "You are not logged in. Please run 'argilla login' to login to an Argilla server.",
            title="Not logged in",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1)

    try:
        init()
    except Exception as e:
        echo_in_panel(
            "The Argilla Server you are logged in is not available or not responding. Please make sure it's running and try again.",
            title="Server not available",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e


def deprecated_database_cmd_callback(ctx: typer.Context) -> None:
    echo_in_panel(
        f"Instead you should run `argilla server database {ctx.invoked_subcommand}`",
        title="Deprecated command",
        title_align="left",
        success=False,
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer

app = typer.Typer(invoke_without_command=True)


@app.callback(help="Logout from an Argilla Server")
def logout(force: bool = typer.Option(False, help="Force the logout even if the server cannot be reached")) -> None:
    from argilla.cli.callback import init_callback
    from argilla.cli.rich import echo_in_panel
    from argilla.client.login import ArgillaCredentials

    if not force:
        init_callback()

    ArgillaCredentials.remove()

    echo_in_panel("Logged out successfully from Argilla server!", title="Logout", title_align="left")


if __name__ == "__main__":
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Optional, Union

from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import Response
from argilla.client.sdk.v1.vectors_settings.models import VectorSettingsModel

if TYPE_CHECKING:
    from uuid import UUID

    import httpx

    from argilla.client.sdk.commons.models import ErrorMessage, HTTPValidationError


def update_vector_settings(
    client: "httpx.Client", id: Union[str, "UUID"], title: str
) -> Response[Union[VectorSettingsModel, "ErrorMessage", "HTTPValidationError"]]:
    """Sends a `PATCH` request to `/api/v1/vector_settings/{id}` endpoint to partially
    update a vector settings in Argilla.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the vector settings to be updated in Argilla.
        title: the new title of the vector settings.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a `VectorSettingsModel`.
    """
    url = f"/api/v1/vectors-settings/{id}"

    body = {"title": title}

    response = client.patch(url=url, json=body)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = VectorSettingsModel.parse_raw(response.content)
        return response_obj

    return handle_response_error(response)


def delete_vector_settings(
    client: "httpx.Client", id: Union[str, "UUID"]
) -> Response[Union[VectorSettingsModel, "ErrorMessage", "HTTPValidationError"]]:
    """Sends a `DELETE` request to `/api/v1/vector_settings/{id}` endpoint to delete
    a vector settings in Argilla.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the vector settings to be deleted in Argilla.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a `VectorSettingsModel`.
    """
    url = f"/api/v1/vectors-settings/{id}"

    response = client.delete(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = VectorSettingsModel.parse_raw(response.content)
        return response_obj

    return handle_response_error(response)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union
from uuid import UUID

import httpx

from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import ErrorMessage, HTTPValidationError, Response
from argilla.client.sdk.v1.suggestions.models import SuggestionModel


def delete_suggestion(
    client: httpx.Client, id: UUID
) -> Response[Union[SuggestionModel, ErrorMessage, HTTPValidationError]]:
    """Sends a DELETE request to `/api/v1/suggestions/{suggestion_id}` endpoint to delete
    a suggestion from Argilla.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the suggestion to be deleted from Argilla.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a `SuggestionModel`.
    """
    url = f"/api/v1/suggestions/{id}"

    response = client.delete(url=url)

    if response.status_code == 200:
        response = Response.from_httpx_response(response)
        response.parsed = SuggestionModel.parse_raw(response.content)
        return response
    return handle_response_error(response)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Tuple, Union
from uuid import UUID

from argilla.client.feedback.schemas.enums import RecordSortField, ResponseStatus, SortOrder
from argilla.pydantic_v1 import BaseModel, Extra, Field, PrivateAttr, StrictInt, StrictStr, conint, validator

if TYPE_CHECKING:
    from argilla.client.feedback.unification import UnifiedValueSchema


class RankingValueSchema(BaseModel):
    """Schema for the `RankingQuestion` response value for a `RankingQuestion`. Note that
    we may have more than one record in the same rank.

    Args:
        value: The value of the record.
        rank: The rank of the record.
    """

    value: StrictStr
    rank: Optional[conint(ge=1)] = None


class ValueSchema(BaseModel):
    """Schema for any `FeedbackRecord` response value.

    Args:
        value: The value of the record.
    """

    value: Union[StrictStr, StrictInt, List[str], List[RankingValueSchema]]


class ResponseSchema(BaseModel):
    """Schema for the `FeedbackRecord` response.

    Args:
        user_id: ID of the user that provided the response. Defaults to None, and is
            automatically fulfilled internally once the question is pushed to Argilla.
        values: Values of the response, should match the questions in the record.
        status: Status of the response. Defaults to `submitted`.

    Examples:
        >>> from argilla.client.feedback.schemas.records import ResponseSchema
        >>> ResponseSchema(
        ...     values={
        ...         "question_1": {"value": "answer_1"},
        ...         "question_2": {"value": "answer_2"},
        ...     }
        ... )
    """

    user_id: Optional[UUID] = None
    values: Union[Dict[str, ValueSchema], None]
    status: ResponseStatus = ResponseStatus.submitted

    class Config:
        extra = Extra.forbid
        validate_assignment = True

    @validator("user_id", always=True)
    def user_id_must_have_value(cls, v):
        if not v:
            warnings.warn(
                "`user_id` not provided, so it will be set to `None`. Which is not an"
                " issue, unless you're planning to log the response in Argilla, as"
                " it will be automatically set to the active `user_id`.",
            )
        return v

    def to_server_payload(self) -> Dict[str, Any]:
        """Method that will be used to create the payload that will be sent to Argilla
        to create a `ResponseSchema` for a `FeedbackRecord`."""
        return {
            # UUID is not json serializable!!!
            "user_id": self.user_id,
            "values": {question_name: value.dict() for question_name, value in self.values.items()}
            if self.values is not None
            else None,
            "status": self.status.value if hasattr(self.status, "value") else self.status,
        }


class SuggestionSchema(BaseModel):
    """Schema for the suggestions for the questions related to the record.

    Args:
        question_name: name of the question in the `FeedbackDataset`.
        type: type of the question. Defaults to None. Possible values are `model` or `human`.
        score: score of the suggestion. Defaults to None.
        value: value of the suggestion, which should match the type of the question.
        agent: agent that generated the suggestion. Defaults to None.

    Examples:
        >>> from argilla.client.feedback.schemas.records import SuggestionSchema
        >>> SuggestionSchema(
        ...     question_name="question-1",
        ...     type="model",
        ...     score=0.9,
        ...     value="This is the first suggestion",
        ...     agent="agent-1",
        ... )
    """

    question_name: str
    type: Optional[Literal["model", "human"]] = None
    score: Optional[float] = None
    value: Any
    agent: Optional[str] = None

    class Config:
        extra = Extra.forbid
        validate_assignment = True

    def to_server_payload(self, question_name_to_id: Dict[str, UUID]) -> Dict[str, Any]:
        """Method that will be used to create the payload that will be sent to Argilla
        to create a `SuggestionSchema` for a `FeedbackRecord`."""
        payload = {}
        payload["question_id"] = str(question_name_to_id[self.question_name])
        payload["value"] = self.value
        if self.type:
            payload["type"] = self.type
        if self.score:
            payload["score"] = self.score
        if self.agent:
            payload["agent"] = self.agent
        return payload


class FeedbackRecord(BaseModel):
    """Schema for the records of a `FeedbackDataset`.

    Args:
        fields: Fields that match the `FeedbackDataset` defined fields. So this attribute
            contains the actual information shown in the UI for each record, being the
            record itself.
        metadata: Metadata to be included to enrich the information for a given record.
            Note that the metadata is not shown in the UI so you'll just be able to see
            that programmatically after pulling the records. Defaults to None.
        responses: Responses given by either the current user, or one or a collection of
            users that must exist in Argilla. Each response corresponds to one of the
            `FeedbackDataset` questions, so the values should match the question type.
            Defaults to None.
        suggestions: A list of `SuggestionSchema` that contains the suggestions
            for the current record. Every suggestion is linked to only one
            question. Defaults to an empty list.
        external_id: The external ID of the record, which means that the user can
            specify this ID to identify the record no matter what the Argilla ID is.
            Defaults to None.

    Examples:
        >>> from argilla.client.feedback.schemas.records import FeedbackRecord, ResponseSchema, SuggestionSchema, ValueSchema
        >>> FeedbackRecord(
        ...     fields={"text": "This is the first record", "label": "positive"},
        ...     metadata={"first": True, "nested": {"more": "stuff"}},
        ...     responses=[ # optional
        ...         ResponseSchema(
        ...             user_id="user-1",
        ...             values={
        ...                 "question-1": ValueSchema(value="This is the first answer"),
        ...                 "question-2": ValueSchema(value=5),
        ...             },
        ...             status="submitted",
        ...         ),
        ...     ],
        ...     suggestions=[ # optional
        ...         SuggestionSchema(
        ...            question_name="question-1",
        ...            type="model",
        ...            score=0.9,
        ...            value="This is the first suggestion",
        ...            agent="agent-1",
        ...         ),
        ...     external_id="entry-1",
        ... )

    """

    fields: Dict[str, Union[str, None]]
    metadata: Dict[str, Any] = Field(default_factory=dict)
    vectors: Dict[str, List[float]] = Field(default_factory=dict)
    responses: List[ResponseSchema] = Field(default_factory=list)
    suggestions: Union[Tuple[SuggestionSchema], List[SuggestionSchema]] = Field(default_factory=tuple)
    external_id: Optional[str] = None

    _unified_responses: Optional[Dict[str, List["UnifiedValueSchema"]]] = PrivateAttr(default_factory=dict)

    class Config:
        extra = Extra.forbid
        validate_assignment = True

    @validator("suggestions", always=True)
    def normalize_suggestions(cls, values: Any) -> Tuple:
        if not isinstance(values, tuple):
            return tuple([v for v in values])
        return values

    @property
    def unified_responses(self) -> Optional[Dict[str, List["UnifiedValueSchema"]]]:
        """Property that returns the unified responses for the record."""
        return self._unified_responses

    def update(
        self, suggestions: Union[SuggestionSchema, List[SuggestionSchema], Dict[str, Any], List[Dict[str, Any]]]
    ) -> None:
        if isinstance(suggestions, (dict, SuggestionSchema)):
            suggestions = [suggestions]
        parsed_suggestions = []
        for suggestion in suggestions:
            if not isinstance(suggestion, SuggestionSchema):
                suggestion = SuggestionSchema(**suggestion)
            parsed_suggestions.append(suggestion)

        suggestions_dict = {suggestion.question_name: suggestion for suggestion in self.suggestions}
        for suggestion in parsed_suggestions:
            if suggestion.question_name in suggestions_dict:
                warnings.warn(
                    f"A suggestion for question `{suggestion.question_name}` has already"
                    " been provided, so the provided suggestion will overwrite it.",
                    category=UserWarning,
                )
            suggestions_dict[suggestion.question_name] = suggestion

        self.__dict__["suggestions"] = tuple(suggestions_dict.values())

    def to_server_payload(self, question_name_to_id: Optional[Dict[str, UUID]] = None) -> Dict[str, Any]:
        """Method that will be used to create the payload that will be sent to Argilla
        to create a `FeedbackRecord` in the `FeedbackDataset`.
        """
        payload = {}
        payload["fields"] = {key: value for key, value in self.fields.items() if value is not None}
        if self.responses:
            payload["responses"] = [response.to_server_payload() for response in self.responses]
        if question_name_to_id:
            payload["suggestions"] = [
                suggestion.to_server_payload(question_name_to_id) for suggestion in self.suggestions
            ]

        if self.vectors:
            payload["vectors"] = self.vectors
        if self.metadata:
            payload["metadata"] = self.metadata
        if self.external_id:
            payload["external_id"] = self.external_id
        return payload


class SortBy(BaseModel):
    field: Union[str, RecordSortField]
    order: Union[str, SortOrder] = SortOrder.asc

    @validator("field", pre=True)
    def check_field_name(cls, field: Union[str, RecordSortField]) -> Union[str, RecordSortField]:
        try:
            return RecordSortField(field)
        except ValueError:
            if field.startswith("metadata."):
                return field
            else:
                raise ValueError(
                    f"{field} is not a valid field name. Supported fields are: {RecordSortField} or metadata.*"
                )

    @validator("order")
    def check_order(cls, order):
        return SortOrder(order)

    @property
    def is_metadata_field(self) -> bool:
        """Returns whether the field is a metadata field."""
        return self.field.startswith("metadata.")

    @property
    def metadata_name(self) -> Optional[str]:
        """Returns the name of the metadata field."""
        if self.field.startswith("metadata."):
            return self.field.split("metadata.")[1]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional, Union

import typer


def list_users(
    workspace: Optional[str] = typer.Option(None, help="Filter users by workspace"),
    role: Optional[str] = typer.Option(None, help="Filter users by role"),
) -> None:
    from rich.console import Console

    from argilla.cli.rich import echo_in_panel, get_argilla_themed_table
    from argilla.client.sdk.v1.workspaces.models import WorkspaceModel
    from argilla.client.users import User
    from argilla.client.workspaces import Workspace

    def build_workspaces_text(workspaces: Union[List[str], List[WorkspaceModel], None]) -> str:
        text = ""
        if not workspaces:
            return text
        for i, workspace in enumerate(workspaces):
            workspace_name = workspace.name if isinstance(workspace, WorkspaceModel) else workspace
            text += f"• {workspace_name}"
            if i < len(workspaces) - 1:
                text += "\n"
        return text

    try:
        if workspace is not None:
            try:
                users = Workspace.from_name(workspace).users
            except ValueError as e:
                echo_in_panel(
                    f"Workspace with name={workspace} does not exist.",
                    title="Workspace not found",
                    title_align="left",
                    success=False,
                )
                raise typer.Exit(code=1) from e
        else:
            users = User.list()
    except RuntimeError as e:
        echo_in_panel(
            "An unexpected error occurred when trying to retrieve the list of users from the Argilla server.",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    table = get_argilla_themed_table(title="Users", show_lines=True)
    for column in (
        "ID",
        "Username",
        "Role",
        "First name",
        "Last name",
        "Workspaces",
        "Creation Date",
        "Last Updated Date",
    ):
        table.add_column(column)

    for user in users:
        if role is not None and user.role != role:
            continue
        table.add_row(
            str(user.id),
            user.username,
            user.role,
            user.first_name,
            user.last_name,
            build_workspaces_text(user.workspaces),
            user.inserted_at.isoformat(sep=" "),
            user.updated_at.isoformat(sep=" "),
        )

    Console().print(table)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from dataclasses import dataclass, field, fields
from inspect import getsource
from pathlib import Path
from platform import python_version
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Literal, Optional, Union

from huggingface_hub import CardData, ModelCard, dataset_info, model_info
from huggingface_hub.utils import yaml_dump

from argilla._version import version
from argilla.client.feedback.training.schemas.base import TRAINING_TASK_MAPPING, TrainingTaskTypes
from argilla.client.models import FRAMEWORK_TO_NAME_MAPPING, Framework
from argilla.training.utils import get_default_args

if TYPE_CHECKING:
    import spacy
    from transformers import PreTrainedTokenizer


TEMPLATE_ARGILLA_MODEL_CARD_PATH = Path(__file__).parent / "argilla_model_template.md"


TEMPLATE_TASK_CALL = "task = TrainingTask.{task_type}({training_task_args})"


YAML_FIELDS = ["language", "license", "tags", "dataset_name", "library_name"]


class ArgillaModelCard(ModelCard):
    """`ArgillaModelCard` has been created similarly to `ModelCard` from
    `huggingface_hub` but with a different template. The template is located at
    `argilla/client/feedback/integrations/huggingface/model_card/argilla_model_template.md`.
    """

    default_template_path = TEMPLATE_ARGILLA_MODEL_CARD_PATH


@dataclass
class FrameworkCardData(CardData):
    """Parent class to generate the variables to add to the ModelCard.

    Each framework will inherit from here and update accordingly.
    """

    # User provided
    language: Optional[Union[str, List[str]]] = None
    license: Optional[str] = None
    model_name: Optional[str] = None
    model_id: Optional[str] = None
    dataset_name: Optional[str] = None
    dataset_id: Optional[str] = None
    tags: Optional[List[str]] = field(default_factory=lambda: ["argilla"])
    model_summary: Optional[str] = None
    model_description: Optional[str] = None
    developers: Optional[str] = None
    shared_by: Optional[str] = None
    model_type: Optional[str] = None
    finetuned_from: Optional[str] = None
    repo: Optional[str] = None

    # Control variables for the templates
    _is_on_huggingface: bool = field(default=False)

    # Obtained internally from each trainer
    framework: Optional[Framework] = None
    train_size: Optional[float] = None
    seed: Optional[int] = None
    framework_kwargs: Dict[str, Any] = field(default_factory=dict)
    task: Optional[TrainingTaskTypes] = None
    output_dir: Optional[str] = None
    version: Dict[str, str] = field(
        default_factory=lambda: {
            "python": python_version(),
            "argilla": version,
        },
        init=False,
    )
    library_name: Optional[str] = None
    # Used to store the arguments passed through `update_config` method. In the case
    # of transformers for example, this corresponds to the `trainer_kwargs`.
    update_config_kwargs: Dict[str, Any] = field(default_factory=lambda: {})

    def __post_init__(self):
        # To decide whether the dataset is loaded from from_huggingface or from_argilla
        if self.dataset_name:
            if is_on_huggingface(self.dataset_name, is_model=False):
                self._is_on_huggingface = True
        self.library_name = FRAMEWORK_TO_NAME_MAPPING[self.framework.value]
        self.task_type = TRAINING_TASK_MAPPING[type(self.task)]

    def _trainer_task__repr__(self) -> str:
        """Generates the creation of the `TrainingTask*` call.

        Returns:
            Representation of the training task creation as a str.
        """
        pass

    def _predict__repr__(self) -> str:
        """Generates the call to the `predict` method, for the models that implement it, or
        the underlying library implementation.

        Returns:
            A sample call to the predict method according to the type of model.
        """
        pass

    def _update_config__repr__(self) -> str:
        """Generates the call to the `update_config` method, for the models that implement it.
        The arguments passed by the user to `update_config` are the difference between what
        the model contains in the `trainer_kwargs`, which are different across frameworks, (not
        only the internal variables but the the attribute name), and what it's internally
        generated by the `init_training_args` method.

        Returns:
            A sample call to the predict method according to the type of model.
        """
        pass

    def _to_dict(self) -> Dict[str, str]:
        """Write this method to insert variables pertaining to a special framework only."""
        return {}

    def to_dict(self) -> Dict[str, Any]:
        """Main method to generate the variables that will be written in the model card."""
        default_kwargs = {field.name: getattr(self, field.name) for field in fields(self)}

        kwargs = {
            "framework": self.framework.value,
            "trainer_task_call": self._trainer_task__repr__(),
            "predict_call": self._predict__repr__(),
        }

        if self.framework_kwargs:
            kwargs["framework_kwargs"] = str(self.framework_kwargs)

        if update_config_kwargs := self._update_config__repr__():
            kwargs["update_config_call"] = update_config_kwargs

        if extra_kwargs := self._to_dict():
            kwargs.update(**extra_kwargs)

        return {**default_kwargs, **kwargs}

    def to_yaml(self, line_break=None) -> str:
        return yaml_dump(
            {key: value for key, value in self.to_dict().items() if key in YAML_FIELDS and value is not None},
            sort_keys=False,
            line_break=line_break,
        ).strip()


@dataclass
class SpacyModelCardDataBase(FrameworkCardData):
    lang: Optional["spacy.Language"] = None
    gpu_id: Optional[int] = -1
    optimize: Literal["efficiency", "accuracy"] = "efficiency"
    pipeline: List[str] = field(default_factory=lambda: ["ner"])

    def _trainer_task__repr__(self) -> str:
        task_call = ""
        if formatting_func := self.task.formatting_func:
            task_call += getsource(formatting_func) + "\n"
            training_task_args = f"formatting_func={formatting_func.__name__}"
        else:
            text = f'dataset.field_by_name("{self.task.text.name}")'
            training_task_args = f'text={text}, label=dataset.question_by_name("{self.task.label.question.name}")'
        return task_call + TEMPLATE_TASK_CALL.format(task_type=self.task_type, training_task_args=training_task_args)

    def _predict__repr__(self) -> str:
        return 'trainer.predict("This is awesome!")'

    def _to_dict(self) -> Dict[str, str]:
        return {"gpu_id": self.gpu_id, "lang": self.lang, "optimize": self.optimize}


@dataclass
class SpacyModelCardData(SpacyModelCardDataBase):
    framework: Framework = Framework("spacy")
    freeze_tok2vec: bool = False

    def _to_dict(self) -> Dict[str, str]:
        kwargs = super()._to_dict()
        # Only add this variable if is different from the default
        if freeze_tok2vec := self.freeze_tok2vec:
            kwargs.update({"freeze_tok2vec": freeze_tok2vec})
        return kwargs

    def _update_config__repr__(self) -> Optional[str]:
        return


@dataclass
class SpacyTransformersModelCardData(SpacyModelCardDataBase):
    framework: Framework = Framework("spacy-transformers")
    update_transformer: bool = True

    def _to_dict(self) -> Dict[str, str]:
        kwargs = super()._to_dict()
        if update_transformer := not self.update_transformer:
            kwargs.update({"update_transformer": update_transformer})
        return kwargs

    def _update_config__repr__(self) -> str:
        return


@dataclass
class TransformersModelCardDataBase(FrameworkCardData):
    tokenizer: "PreTrainedTokenizer" = ""

    def _trainer_task__repr__(self) -> str:
        task_call = ""
        if formatting_func := self.task.formatting_func:
            task_call += getsource(formatting_func) + "\n"
            training_task_args = f"formatting_func={formatting_func.__name__}"
        else:
            text = f'dataset.field_by_name("{self.task.text.name}")'
            training_task_args = f'text={text}, label=dataset.question_by_name("{self.task.label.question.name}")'
        return task_call + TEMPLATE_TASK_CALL.format(task_type=self.task_type, training_task_args=training_task_args)

    def _predict__repr__(self) -> str:
        return 'trainer.predict("This is awesome!")'

    def _to_dict(self) -> Dict[str, str]:
        return {"tokenizer": self.tokenizer}


@dataclass
class TransformersModelCardData(TransformersModelCardDataBase):
    framework: Framework = Framework("transformers")

    def _trainer_task__repr__(self) -> str:
        task_call = ""
        if formatting_func := self.task.formatting_func:
            task_call += getsource(formatting_func) + "\n"
            training_task_args = f"formatting_func={formatting_func.__name__}"
        else:
            if self.task_type == "for_text_classification":
                training_task_args = (
                    f'text=dataset.field_by_name("{self.task.text.name}"), '
                    f'label=dataset.question_by_name("{self.task.label.question.name}")'
                )
            elif self.task_type == "for_question_answering":
                training_task_args = (
                    f'question=dataset.field_by_name("{self.task.question.name}"), '
                    f'context=dataset.field_by_name("{self.task.context.name}"), '
                    f'answer=dataset.question_by_name("{self.task.answer.name}")'
                )
            else:
                raise NotImplementedError(f"Transformer doesn't have this `task_type` implemented: `{self.task_type}`")

        return task_call + TEMPLATE_TASK_CALL.format(task_type=self.task_type, training_task_args=training_task_args)

    def _predict__repr__(self) -> str:
        if self.task_type == "for_text_classification":
            return super()._predict__repr__()
        elif self.task_type == "for_question_answering":
            return dedent(
                f"""\
                # This type of model has no `predict` method implemented from argilla, but can be done using the underlying library

                from transformers import pipeline

                qa_model = pipeline("question-answering", model="{self.output_dir}")
                question = "Where do I live?"
                context = "My name is Merve and I live in İstanbul."
                qa_model(question = question, context = context)"""
            )
        else:
            raise NotImplementedError(f"`task_type` not implemented: `{self.task_type}`")

    def _update_config__repr__(self) -> Optional[str]:
        from transformers import TrainingArguments

        base_kwargs = get_default_args(TrainingArguments.__init__)

        if updated_args := _updated_arguments(base_kwargs, self.update_config_kwargs):
            return _update_config__repr__(updated_args)


@dataclass
class SetFitModelCardData(TransformersModelCardDataBase):
    framework: Framework = Framework("setfit")
    tags: Optional[List[str]] = field(default_factory=lambda: ["text-classification", "setfit", "argilla"])

    def _update_config__repr__(self) -> Optional[str]:
        from setfit import SetFitModel, SetFitTrainer

        setfit_model_kwargs = get_default_args(SetFitModel._from_pretrained)
        setfit_model_kwargs.update(get_default_args(SetFitModel.from_pretrained))

        setfit_trainer_kwargs = get_default_args(SetFitTrainer.__init__)

        # The following arguments are set by default internally, don't need to be shown.
        self.update_config_kwargs.pop("model_id", None)
        self.update_config_kwargs.pop("revision", None)
        self.update_config_kwargs.pop("pretrained_model_name_or_path", None)
        self.update_config_kwargs.pop("multi_target_strategy", None)
        self.update_config_kwargs.pop("device", None)
        self.update_config_kwargs.pop("column_mapping", None)
        self.update_config_kwargs.pop("train_dataset", None)
        self.update_config_kwargs.pop("eval_dataset", None)
        self.update_config_kwargs.pop("model", None)

        base_kwargs = {**setfit_model_kwargs, **setfit_trainer_kwargs}

        if updated_args := _updated_arguments(base_kwargs, self.update_config_kwargs):
            return _update_config__repr__(updated_args)


@dataclass
class SpanMarkerModelCardData(TransformersModelCardDataBase):
    # Not implemented, once we have the FeedbackDataset ready for this model it should be aligned.
    framework: Framework = Framework("span_marker")
    tags: Optional[List[str]] = field(
        default_factory=lambda: [
            "span-marker",
            "token-classification",
            "ner",
            "named-entity-recognition",
        ]
    )


@dataclass
class PeftModelCardData(TransformersModelCardDataBase):
    framework: Framework = Framework("peft")

    def _update_config__repr__(self) -> Optional[str]:
        base_kwargs = {
            "r": 8,
            "target_modules": None,
            "lora_alpha": 16,
            "lora_dropout": 0.1,
            "fan_in_fan_out": False,
            "bias": "none",
            "inference_mode": False,
            "modules_to_save": None,
            "init_lora_weights": True,
        }

        self.update_config_kwargs.pop("task_type", None)

        if updated_args := _updated_arguments(base_kwargs, self.update_config_kwargs):
            return _update_config__repr__(updated_args)


@dataclass
class OpenAIModelCardData(FrameworkCardData):
    framework: Framework = Framework("openai")

    def _trainer_task__repr__(self) -> str:
        return _formatting_func_call(self.task.formatting_func, self.task_type)

    def _predict__repr__(self) -> str:
        return dedent(
            """\
            # After training we can use the model from the openai framework, you can take a look at their docs in order to use the model
            import openai

            completion = openai.ChatCompletion.create(
                model="ft:gpt-3.5-turbo:my-org:custom_suffix:id",
                messages=[
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": "Hello!"}
                ]
            )
            """
        )


@dataclass
class TRLModelCardData(FrameworkCardData):
    framework: Framework = Framework("trl")

    def _trainer_task__repr__(self) -> str:
        return _formatting_func_call(self.task.formatting_func, self.task_type)

    def _predict__repr__(self) -> str:
        predict_call = "# This type of model has no `predict` method implemented from argilla, but can be done using the underlying library\n"
        if self.task_type == "for_supervised_fine_tuning":
            return predict_call + dedent(
                f"""\
                from transformers import GenerationConfig, AutoTokenizer, GPT2LMHeadModel

                def generate(model_id: str, instruction: str, context: str = "") -> str:
                    model = GPT2LMHeadModel.from_pretrained(model_id)
                    tokenizer = AutoTokenizer.from_pretrained(model_id)

                    inputs = template.format(
                        instruction=instruction,
                        context=context,
                        response="",
                    ).strip()

                    encoding = tokenizer([inputs], return_tensors="pt")
                    outputs = model.generate(
                        **encoding,
                        generation_config=GenerationConfig(
                            max_new_tokens=32,
                            min_new_tokens=12,
                            pad_token_id=tokenizer.pad_token_id,
                            eos_token_id=tokenizer.eos_token_id,
                        ),
                    )
                    return tokenizer.decode(outputs[0])

                generate("{self.output_dir.replace('"', '')}", "Is a toad a frog?")"""
            )
        elif self.task_type == "for_reward_modeling":
            return predict_call + dedent(
                f"""\
                from transformers import AutoTokenizer, AutoModelForSequenceClassification
                import torch

                model = AutoModelForSequenceClassification.from_pretrained("{self.output_dir.replace('"', "")}")
                tokenizer = AutoTokenizer.from_pretrained("{self.output_dir.replace('"', "")}")

                def get_score(model, tokenizer, text):
                    # Tokenize the input sequences
                    inputs = tokenizer(text, truncation=True, padding="max_length", max_length=512, return_tensors="pt")

                    # Perform forward pass
                    with torch.no_grad():
                        outputs = model(**inputs)

                    # Extract the logits
                    return outputs.logits[0, 0].item()

                # Example usage
                example = template.format(instruction="your prompt", context="your context", response="response")

                score = get_score(model, tokenizer, example)
                print(score)"""
            )
        elif (self.task_type == "for_proximal_policy_optimization") or (
            self.task_type == "for_direct_preference_optimization"
        ):
            return predict_call + dedent(
                f"""\
                from transformers import AutoModelForCausalLM, AutoTokenizer

                model = AutoModelForCausalLM.from_pretrained("{self.output_dir.replace('"', "")}")
                tokenizer = AutoTokenizer.from_pretrained("{self.output_dir.replace('"', "")}")
                tokenizer.pad_token = tokenizer.eos_token

                inputs = template.format(
                    instruction="your prompt",
                    context="your context",
                    response=""
                ).strip()
                encoding = tokenizer([inputs], return_tensors="pt")
                outputs = model.generate(**encoding, max_new_tokens=30)
                output_text = tokenizer.decode(outputs[0])
                print(output_text)"""
            )
        else:
            raise NotImplementedError(f"Transformer doesn't have this `task_type` implemented: `{self.task_type}`")

    def _update_config__repr__(self) -> Optional[str]:
        if self.task_type == "for_proximal_policy_optimization":
            # Similar to what happens with spacy, the current implementation
            # doesn't render appropriately, for the moment let for the user to be written by hand.
            return

        else:
            base_kwargs = {
                # Let evaluation_strategy as if the eval dataset was passed by default
                "evaluation_strategy": "epoch",
                "logging_steps": 1,
                "num_train_epochs": 1,
            }

        if updated_args := _updated_arguments(base_kwargs, self.update_config_kwargs):
            return _update_config__repr__(updated_args)


@dataclass
class SentenceTransformerCardData(FrameworkCardData):
    framework: Framework = Framework("sentence-transformers")
    tags: Optional[List[str]] = field(
        default_factory=lambda: ["sentence-similarity", "sentence-transformers", "argilla"]
    )
    cross_encoder: bool = False
    # Used to gather internally the arguments passed to `update_config`
    trainer_cls: Optional[Callable] = None

    def _trainer_task__repr__(self) -> str:
        task_call = ""
        if formatting_func := self.task.formatting_func:
            task_call += getsource(formatting_func) + "\n"
            training_task_args = f"formatting_func={formatting_func.__name__}"
        else:
            texts = ", ".join([f'dataset.field_by_name("{text.name}")' for text in self.task.texts])
            training_task_args = f"texts=[{texts}]{f', label=dataset.question_by_name({self.task.label.question.name})' if self.task.label else ''}"
        return task_call + TEMPLATE_TASK_CALL.format(task_type=self.task_type, training_task_args=training_task_args)

    def _predict__repr__(self) -> str:
        return dedent(
            """\
            trainer.predict(
                [
                    ["Machine learning is so easy.", "Deep learning is so straightforward."],
                    ["Machine learning is so easy.", "This is so difficult, like rocket science."],
                    ["Machine learning is so easy.", "I can't believe how much I struggled with this."]
                ]
            )"""
        )

    def _to_dict(self) -> Dict[str, str]:
        if cross_encoder := self.cross_encoder:
            return {"cross_encoder": cross_encoder}
        return {}

    def _update_config__repr__(self) -> Optional[str]:
        base_kwargs = {
            # model_kwargs
            **get_default_args(self.trainer_cls.__init__),
            # trainer_kwargs
            **get_default_args(self.trainer_cls.fit),
            # data_kwargs
            "batch_size": 32,
            "dataset_type": None,
        }

        if updated_args := _updated_arguments(base_kwargs, self.update_config_kwargs):
            return _update_config__repr__(updated_args)


def _formatting_func_call(formatting_func: Callable, task_type: str) -> str:
    """Helper function to extract the code for the task call.

    Args:
        formatting_func: Function used to prepare the dataset for training.
        task_type: Method called to prepare the dataset for training.

    Returns:
        formatting_func_call
    """
    task_call = getsource(formatting_func) + "\n"
    training_task_args = f"formatting_func={formatting_func.__name__}"
    return task_call + TEMPLATE_TASK_CALL.format(task_type=task_type, training_task_args=training_task_args)


def _updated_arguments(base_kwargs: Dict[str, Any], current_kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Helper function to determine the arguments the user has given through the `update_config` method.

    It does so by obtaining the difference between the `current_kwargs` and the `base_kwargs`
    (the one used by default in the model).

    The arguments can contain nested dicts (which are unhashable). Nested dicts (only one level of depth) are first
    transformed to tuples of their items to check for differences in the values, and then transformed back.
    Instantiated classes and type objects (a class not instantiated for example) are transformed to their name.

    Args:
        base_kwargs: default arguments.
        current_kwargs: arguments registered in the model after training.

    Returns:
        user_kwargs: User provided kwargs.
    """

    base_kwargs_ = _prepare_dict_for_comparison(base_kwargs)
    current_kwargs_ = _prepare_dict_for_comparison(current_kwargs)

    set1 = set(base_kwargs_.items())
    set2 = set(current_kwargs_.items())
    user_kwargs = dict(set2.difference(set1))

    return _prepare_dict_for_return(user_kwargs)


def _prepare_dict_for_comparison(d):
    # Transforms the nested lists/dicts to tuples and prepares
    # them to be recovered after finding possible differences.
    # If a "type" is found, it returns its string representation.
    new_dict = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = list(v.items())
            v.append("__dict")  # Placeholder to create it back
            v = tuple(v)
        elif isinstance(v, list):
            v = [v if isinstance(v, (int, float)) else str(v) for v in v]
            v.append("__list")
            v = tuple(v)
        elif isinstance(v, (int, float, bool)):
            # Left these values as they are
            pass
        else:
            # The remaining cases are either strings (the general case)
            # an instance of a class (in which case we use the str representation)
            # or a class (in the best case in the classes' cases there exists a nice __repr__ method as to be
            # valid for `update_config`, otherwise a better logic must be implemented).
            if isinstance(v, type):
                v = v.__name__
            else:
                v = str(v)

        new_dict[k] = v
    return new_dict


def _prepare_dict_for_return(d):
    # Prepares the dict with the original lists/dicts
    new_dict = {}
    for k, v in d.items():
        if isinstance(v, tuple):
            if "__dict" in v:
                v = list(v)
                v.remove("__dict")
                v = dict(v)
            elif "__list" in v:
                v = list(v)
                v.remove("__list")

        new_dict[k] = v
    return new_dict


def _update_config__repr__(keyword_arguments: Dict[str, Any]) -> str:
    """Creates the call to `update_config` on the model.

    Args:
        keyword_arguments: Arguments given by the user.

    Returns:
        trainer.update_config(...) call.
    """
    return f"\ntrainer.update_config({json.dumps(keyword_arguments, sort_keys=True, indent=4)})\n"


def is_on_huggingface(repo_id: str, is_model: bool = True) -> bool:
    # NOTE: kindly copied from https://github.com/tomaarsen/SpanMarkerNER/blob/main/span_marker/model_card.py
    # Models with more than two 'sections' certainly are not public models
    if len(repo_id.split("/")) > 2:
        return False

    try:
        if is_model:
            model_info(repo_id)
        else:
            dataset_info(repo_id)
        return True
    except:
        # Fetching models can fail for many reasons: Repository not existing, no internet access, HF down, etc.
        return False



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from types import TracebackType
from typing import Any, ContextManager, Dict, Optional, Type

from packaging.version import parse

from argilla.client.apis import AbstractApi
from argilla.client.sdk.client import AuthenticatedClient
from argilla.client.sdk.commons.errors import ApiCompatibilityError
from argilla.pydantic_v1 import BaseModel


class ApiInfo(BaseModel):
    version: Optional[str] = None


class ElasticSearchVersionDetails(BaseModel):
    distribution: Optional[str] = None
    number: str
    build_flavor: Optional[str] = None
    build_type: str
    build_hash: str
    build_date: str
    build_snapshot: bool
    lucene_version: str
    minimum_wire_compatibility_version: str
    minimum_index_compatibility_version: str


class ElasticSearch(BaseModel):
    name: str
    cluster_name: str
    cluster_uuid: str
    version: ElasticSearchVersionDetails
    tagline: str


class ApiStatus(BaseModel):
    version: str
    elasticsearch: ElasticSearch
    mem_info: Dict[str, Any]


class Status(AbstractApi):
    def get_info(self) -> ApiInfo:
        """
        Get the connected API info

        Returns:
            ApiInfo: The API info
        """

        response = self.http_client.get("/api/_info")
        return ApiInfo.parse_obj(response)

    def get_status(self) -> ApiStatus:
        response = self.http_client.get("/api/_status")
        return ApiStatus.parse_obj(response)


class _ApiCompatibilityContextManager(ContextManager):
    def __init__(
        self,
        client: AuthenticatedClient,
        min_version: str,
    ):
        self._min_version = parse(min_version)
        self._status_api = Status(client=client) if client else None

    def __enter__(self):
        if self._status_api:
            api_info = self._status_api.get_info()
            api_version = api_info.version

            api_version = parse(api_version)
            if api_version.is_devrelease:
                api_version = parse(api_version.base_version)
            if not api_version >= self._min_version:
                raise ApiCompatibilityError(str(self._min_version), api_version=api_version)
        pass

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        pass


def api_compatibility(api: AbstractApi, min_version: str):
    """
    Handles problems related to server API compatibility.

    Args:
        api: The api component where apply the api compatibility
        min_version: the minimal version that the argilla client can work with
    """
    return _ApiCompatibilityContextManager(
        min_version=min_version,
        client=api.http_client,
    )



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import random
import uuid
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union

import pandas as pd

from argilla._constants import OPENAI_END_TOKEN, OPENAI_SEPARATOR, OPENAI_WHITESPACE
from argilla.client.apis.datasets import TextClassificationSettings, TokenClassificationSettings
from argilla.client.models import (
    Framework,
    Record,
    Text2TextRecord,
    TextClassificationRecord,
    TokenAttributions,
    TokenClassificationRecord,
)
from argilla.client.sdk.datasets.models import TaskType
from argilla.utils.dependency import require_dependencies, requires_dependencies
from argilla.utils.span_utils import SpanUtils

if TYPE_CHECKING:
    import datasets
    import pandas
    import spacy


class DatasetBase:
    """The Dataset classes are containers for argilla records.

    This is the base class to facilitate the implementation for each record type.

    Args:
        records: A list of argilla records.

    Raises:
        WrongRecordTypeError: When the record type in the provided
            list does not correspond to the dataset type.
    """

    _RECORD_TYPE = None
    # record fields that can hold multiple input columns from a datasets.Dataset or a pandas.DataFrame
    _RECORD_FIELDS_WITH_MULTIPLE_INPUT_COLUMNS = ["inputs", "metadata"]
    _SETTINGS = None

    @classmethod
    def _record_init_args(cls) -> List[str]:
        """
        Helper the returns the field list available for creation of inner records.
        The ``_RECORD_TYPE.__fields__`` will be returned as default
        """
        return [field for field in cls._RECORD_TYPE.__fields__]

    def __init__(self, records: Optional[List[Record]] = None):
        if self._RECORD_TYPE is None:
            raise NotImplementedError("A Dataset implementation has to define a `_RECORD_TYPE`!")

        self._records = records or []
        if self._records:
            self._validate_record_type()

    def _validate_record_type(self):
        """Validates the record type.

        Raises:
            WrongRecordTypeError: When the record type in the provided
                list does not correspond to the dataset type.
        """
        record_types = {type(rec): None for rec in self._records}
        if len(record_types) > 1:
            raise WrongRecordTypeError(
                f"A {type(self).__name__} must only contain"
                f" {self._RECORD_TYPE.__name__}s, but you provided various types:"
                f" {[rt.__name__ for rt in record_types.keys()]}"
            )
        elif next(iter(record_types)) is not self._RECORD_TYPE:
            raise WrongRecordTypeError(
                f"A {type(self).__name__} must only contain"
                f" {self._RECORD_TYPE.__name__}s, but you provided"
                f" {list(record_types.keys())[0].__name__}s."
            )

    def __iter__(self):
        return self._records.__iter__()

    def __getitem__(self, key):
        return self._records[key]

    def __setitem__(self, key, value):
        if type(value) is not self._RECORD_TYPE:
            raise WrongRecordTypeError(
                f"You are only allowed to set a record of type {self._RECORD_TYPE} in"
                f" this dataset, but you provided {type(value)}"
            )
        self._records[key] = value

    def __delitem__(self, key):
        del self._records[key]

    def __len__(self) -> int:
        return len(self._records)

    def __repr__(self):
        return repr(self.to_pandas())

    def __str__(self):
        return repr(self)

    @requires_dependencies("datasets>1.17.0")
    def to_datasets(self) -> "datasets.Dataset":
        """Exports your records to a `datasets.Dataset`.

        Returns:
            A `datasets.Dataset` containing your records.
        """
        import datasets

        ds_dict = self._to_datasets_dict()
        # TODO: THIS FIELD IS ONLY AT CLIENT API LEVEL. NOT SENSE HERE FOR NOW
        if "search_keywords" in ds_dict:
            del ds_dict["search_keywords"]

        try:
            dataset = datasets.Dataset.from_dict(ds_dict)
        # try without metadata, since it is more prone to incompatible structures
        except Exception:
            del ds_dict["metadata"]
            dataset = datasets.Dataset.from_dict(ds_dict)
            warnings.warn(
                "The 'metadata' of the records were removed, since it was incompatible with the 'datasets' format."
            )

        return dataset

    def _to_datasets_dict(self) -> Dict:
        raise NotImplementedError

    @classmethod
    def from_datasets(cls, dataset: "datasets.Dataset", **kwargs) -> "Dataset":
        """Imports records from a `datasets.Dataset`.

        Columns that are not supported are ignored.

        Args:
            dataset: A datasets Dataset from which to import the records.

        Returns:
            The imported records in a argilla Dataset.
        """
        raise NotImplementedError

    @classmethod
    def _prepare_dataset_and_column_mapping(
        cls,
        dataset: "datasets.Dataset",
        column_mapping: Dict[str, Union[str, List[str]]],
    ) -> Tuple["datasets.Dataset", Dict[str, List[str]]]:
        """Renames and removes columns, and extracts the mapping of the columns to be joined.

        Args:
            dataset: A datasets Dataset from which to import the records.
            column_mapping: Mappings from record fields to column names.

        Returns:
            The prepared dataset and a mapping for the columns to be joined
        """
        import datasets

        if isinstance(dataset, datasets.DatasetDict):
            raise ValueError("`datasets.DatasetDict` are not supported. Please, select the dataset split before.")

        # clean column mappings
        column_mapping = {key: val for key, val in column_mapping.items() if val is not None}

        cols_to_be_renamed, cols_to_be_joined = {}, {}
        for field, col in column_mapping.items():
            if field in cls._RECORD_FIELDS_WITH_MULTIPLE_INPUT_COLUMNS:
                cols_to_be_joined[field] = [col] if isinstance(col, str) else col
            else:
                cols_to_be_renamed[col] = field

        dataset = dataset.rename_columns(cols_to_be_renamed)

        dataset = cls._remove_unsupported_columns(
            dataset,
            extra_columns=[col for cols in cols_to_be_joined.values() for col in cols],
        )

        return dataset, cols_to_be_joined

    @classmethod
    def _remove_unsupported_columns(
        cls,
        dataset: "datasets.Dataset",
        extra_columns: List[str],
    ) -> "datasets.Dataset":
        """Helper function to remove unsupported columns from the `datasets.Dataset` following the record type.

        Args:
            dataset: The dataset.
            extra_columns: Extra columns to be kept.

        Returns:
            The dataset with unsupported columns removed.
        """
        not_supported_columns = [
            col for col in dataset.column_names if col not in cls._record_init_args() + extra_columns
        ]

        if not_supported_columns:
            warnings.warn(
                "Following columns are not supported by the"
                f" {cls._RECORD_TYPE.__name__} model and are ignored:"
                f" {not_supported_columns}"
            )
            dataset = dataset.remove_columns(not_supported_columns)

        return dataset

    @staticmethod
    def _join_datasets_columns_and_delete(row: Dict[str, Any], columns: List[str]) -> Dict[str, Any]:
        """Joins columns of a `datasets.Dataset` row into a dict, and deletes the single columns.

        Updates the ``row`` dictionary!

        Args:
            row: A row of a `datasets.Dataset`
            columns: Name of the columns to be joined and deleted from the row.

        Returns:
            A dict containing the columns and its values.
        """
        joined_cols = {}
        for col in columns:
            joined_cols[col] = row[col]
            del row[col]

        return joined_cols

    @staticmethod
    def _parse_datasets_column_with_classlabel(
        column_value: Union[str, List[str], int, List[int]],
        feature: Optional[Any],
    ) -> Optional[Union[str, List[str], int, List[int]]]:
        """Helper function to parse a datasets.Dataset column with a potential ClassLabel feature.

        Args:
            column_value: The value from the datasets Dataset column.
            feature: The feature of the annotation column to optionally convert ints to strs.

        Returns:
            The column value optionally converted to str, or None if the conversion fails.
        """
        import datasets

        # extract ClassLabel feature
        if isinstance(feature, list):
            feature = feature[0]
        if isinstance(feature, datasets.Sequence):
            feature = feature.feature
        if not isinstance(feature, datasets.ClassLabel):
            feature = None

        if feature is None:
            return column_value

        try:
            return feature.int2str(column_value)
        # integers don't have to map to the names ...
        # it seems that sometimes -1 is used to denote "no label"
        except ValueError:
            return None

    def to_pandas(self) -> pd.DataFrame:
        """Exports your records to a `pandas.DataFrame`.

        Returns:
            A `datasets.Dataset` containing your records.
        """
        return pd.DataFrame(map(dict, self._records))

    @classmethod
    def from_pandas(cls, dataframe: pd.DataFrame) -> "Dataset":
        """Imports records from a `pandas.DataFrame`.

        Columns that are not supported are ignored.

        Args:
            dataframe: A pandas DataFrame from which to import the records.

        Returns:
            The imported records in a argilla Dataset.
        """
        not_supported_columns = [col for col in dataframe.columns if col not in cls._record_init_args()]
        if not_supported_columns:
            warnings.warn(
                "Following columns are not supported by the"
                f" {cls._RECORD_TYPE.__name__} model and are ignored:"
                f" {not_supported_columns}"
            )
            dataframe = dataframe.drop(columns=not_supported_columns)

        return cls._from_pandas(dataframe)

    @classmethod
    def _from_pandas(cls, dataframe: pd.DataFrame) -> "Dataset":
        """Helper method to create a argilla Dataset from a pandas DataFrame.

        Must be implemented by the child class.

        Args:
            dataframe: A pandas DataFrame

        Returns:
            A argilla Dataset
        """
        raise NotImplementedError

    def prepare_for_training(
        self,
        framework: Union[Framework, str] = "transformers",
        settings: Union[TextClassificationSettings, TokenClassificationSettings] = None,
        lang: Optional["spacy.Language"] = None,
        train_size: Optional[float] = 1,
        test_size: Optional[float] = None,
        seed: Optional[int] = None,
    ) -> Union[
        "datasets.Dataset",
        "spacy.tokens.DocBin",
        Tuple["spacy.tokens.DocBin", "spacy.tokens.DocBin"],
        Tuple["pandas.DataFrame", "pandas.DataFrame"],
    ]:
        """Prepares the dataset for training.

        This will return a ``datasets.Dataset`` with all columns returned by ``to_datasets`` method
        and an additional  *ner_tags* column:
            - Records without an annotation are removed.
            - The *ner_tags* column corresponds to the iob tags sequences for annotations of the records
            - The iob tags are transformed to integers.

        Args:
            framework: A string|enum specifying the framework for the training.
                "transformers" and "spacy" are currently supported. Default: `transformers`
            lang: The spacy nlp Language pipeline used to process the dataset. (Only for spacy framework)
            train_size: The size of the training set. If float, should be between 0.0 and 1.0 and represent the
                proportion of the dataset to include in the train split.
            test_size: The size of the test set. If float, should be between 0.0 and 1.0 and represent the
                proportion of the dataset to include in the test split.
            seed: Random state.

        Returns:
            A datasets Dataset with a *ner_tags* or a *label* column and and several *inputs* columns.
            returned by ``to_datasets`` for "transformers" framework or a spaCy DocBin for "spacy" framework.

        Examples:
            >>> import argilla as rg
            >>> rb_dataset = rg.DatasetForTokenClassification([
            ...     rg.TokenClassificationRecord(
            ...         text="The text",
            ...         tokens=["The", "text"],
            ...         annotation=[("TAG", 0, 2)],
            ...     )
            ... ])
            >>> rb_dataset.prepare_for_training().features
            {'text': Value(dtype='string'),
             'tokens': Sequence(feature=Value(dtype='string'), length=-1),
             'prediction': Value(dtype='null'),
             'prediction_agent': Value(dtype='null'),
             'annotation': [{'end': Value(dtype='int64'),
               'label': Value(dtype='string'),
               'start': Value(dtype='int64')}],
             'annotation_agent': Value(dtype='null'),
             'id': Value(dtype='null'),
             'metadata': Value(dtype='null'),
             'status': Value(dtype='string'),
             'event_timestamp': Value(dtype='null'),
             'metrics': Value(dtype='null'),
             'ner_tags': [ClassLabel(num_classes=3, names=['O', 'B-TAG', 'I-TAG'])]}

            >>> import argilla as rg
            >>> rb_dataset = rg.DatasetForTextClassification([
            ...     rg.TextClassificationRecord(
            ...         inputs={"header": "my header", "content": "my content"},
            ...         annotation="SPAM",
            ...     )
            ... ])
            >>> rb_dataset.prepare_for_training().features
            {'header': Value(dtype='string'),
             'content': Value(dtype='string'),
             'label': ClassLabel(num_classes=1, names=['SPAM'])}

        """
        if self._RECORD_TYPE == TextClassificationRecord:
            if settings is None:
                self._SETTINGS = self._infer_settings_from_records()
            elif isinstance(settings, TextClassificationSettings):
                self._SETTINGS = settings
            else:
                raise ValueError("settings must be TextClassificationSettings for TextClassificationRecord")
        elif self._RECORD_TYPE == TokenClassificationRecord:
            if settings is None:
                self._SETTINGS = self._infer_settings_from_records()
            elif isinstance(settings, TokenClassificationSettings):
                self._SETTINGS = settings
            else:
                raise ValueError("settings must be TokenClassificationSettings for TokenClassificationRecord")
        else:
            self._SETTINGS = settings

        if train_size is None:
            train_size = 1
        if test_size is None:
            test_size = 1 - train_size

        # check if all numbers are larger than 0
        assert [abs(train_size), abs(test_size)] == [train_size, test_size], ValueError(
            "`train_size` and `test_size` must be larger than 0."
        )

        # check if train sizes sum up to 1
        assert (train_size + test_size) == 1, ValueError("`train_size` and `test_size` must sum to 1.")

        # check for annotations
        assert any([rec.annotation for rec in self._records]), ValueError("Dataset has no annotations.")

        if test_size == 0:
            test_size = None

        # shuffle records
        shuffled_records = self._records.copy()
        seed = seed or random.randint(42, 1984)
        random.Random(seed).shuffle(shuffled_records)

        # turn the string into a Framework instance and trigger error if str is not valid
        if isinstance(framework, str):
            framework = Framework(framework)

        # prepare for training for the right method
        if framework in [
            Framework.TRANSFORMERS,
            Framework.SETFIT,
            Framework.SPAN_MARKER,
            Framework.PEFT,
        ]:
            return self._prepare_for_training_with_transformers(train_size=train_size, test_size=test_size, seed=seed)
        elif framework in [Framework.SPACY, Framework.SPACY_TRANSFORMERS] and lang is None:
            raise ValueError(
                "Please provide a `spaCy` language model to prepare the dataset for"
                " training with the `spaCy`/`spaCy-transformers` framework."
            )
        elif framework in [Framework.SPACY, Framework.SPACY_TRANSFORMERS, Framework.SPARK_NLP, Framework.OPENAI]:
            if train_size and test_size:
                require_dependencies("scikit-learn")
                from sklearn.model_selection import train_test_split

                records_train, records_test = train_test_split(
                    shuffled_records,
                    train_size=train_size,
                    shuffle=False,
                    random_state=seed,
                )
                if framework in [Framework.SPACY, Framework.SPACY_TRANSFORMERS]:
                    train_docbin = self._prepare_for_training_with_spacy(nlp=lang, records=records_train)
                    test_docbin = self._prepare_for_training_with_spacy(nlp=lang, records=records_test)
                    return train_docbin, test_docbin
                elif framework is Framework.SPARK_NLP:
                    train_df = self._prepare_for_training_with_spark_nlp(records_train)
                    test_df = self._prepare_for_training_with_spark_nlp(records_test)
                    return train_df, test_df
                else:
                    train_jsonl = self._prepare_for_training_with_openai(records=records_train)
                    test_jsonl = self._prepare_for_training_with_openai(records=records_test)
                    return train_jsonl, test_jsonl
            else:
                if framework in [Framework.SPACY, Framework.SPACY_TRANSFORMERS]:
                    return self._prepare_for_training_with_spacy(nlp=lang, records=shuffled_records)
                elif framework is Framework.SPARK_NLP:
                    return self._prepare_for_training_with_spark_nlp(records=shuffled_records)
                elif framework is Framework.OPENAI:
                    return self._prepare_for_training_with_openai(records=shuffled_records)
                else:
                    raise NotImplementedError(
                        f"Framework {framework} is not supported. Choose from: {[e.value for e in Framework]}"
                    )
        else:
            raise NotImplementedError(
                f"Framework {framework} is not supported. Choose from: {[e.value for e in Framework]}"
            )

    @requires_dependencies("spacy")
    def _prepare_for_training_with_spacy(
        self, **kwargs
    ) -> Union["spacy.token.DocBin", Tuple["spacy.token.DocBin", "spacy.token.DocBin"]]:
        """Prepares the dataset for training using the "spacy" framework.

        Args:
            **kwargs: Specific to the task of the dataset.

        Returns:
            A spacy.token.DocBin.
        """

        raise NotImplementedError

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_transformers(self, **kwargs) -> "datasets.Dataset":
        """Prepares the dataset for training using the "transformers" framework.

        Args:
            **kwargs: Specific to the task of the dataset.

        Returns:
            A datasets Dataset.
        """

        raise NotImplementedError

    def _prepare_for_training_with_spark_nlp(self, **kwargs) -> "datasets.Dataset":
        """Prepares the dataset for training using the "spark-nlp" framework.

        Args:
            **kwargs: Specific to the task of the dataset.

        Returns:
            A pd.DataFrame.
        """

        raise NotImplementedError

    def _prepare_for_training_with_openai(self, **kwargs) -> "dict":
        """Prepares the dataset for training using the "openai" framework.

        Args:
            **kwargs: Specific to the task of the dataset.

        Returns:
            A pd.DataFrame.
        """

        raise NotImplementedError

    def _infer_settings_from_records(self) -> Union[TokenClassificationSettings, TextClassificationSettings]:
        """Infer the settings from the records.

        Returns:
            A Settings object.
        """

        raise NotImplementedError


def _prepend_docstring(record_type: Type[Record]):
    docstring = f"""This Dataset contains {record_type.__name__} records.

    It allows you to export/import records into/from different formats,
    loop over the records, and access them by index.

    Args:
        records: A list of `{record_type.__name__}`s.

    Raises:
        WrongRecordTypeError: When the record type in the provided
            list does not correspond to the dataset type.
    """

    def docstring_decorator(cls):
        cls.__doc__ = docstring + (cls.__doc__ or "")
        return cls

    return docstring_decorator


@_prepend_docstring(TextClassificationRecord)
class DatasetForTextClassification(DatasetBase):
    """
    Examples:
        >>> # Import/export records:
        >>> import argilla as rg
        >>> dataset = rg.DatasetForTextClassification.from_pandas(my_dataframe)
        >>> dataset.to_datasets()
        >>>
        >>> # Looping over the dataset:
        >>> for record in dataset:
        ...     print(record)
        >>>
        >>> # Passing in a list of records:
        >>> records = [
        ...     rg.TextClassificationRecord(text="example"),
        ...     rg.TextClassificationRecord(text="another example"),
        ... ]
        >>> dataset = rg.DatasetForTextClassification(records)
        >>> assert len(dataset) == 2
        >>>
        >>> # Indexing into the dataset:
        >>> dataset[0]
        ... rg.TextClassificationRecord(text="example")
        >>> dataset[0] = rg.TextClassificationRecord(text="replaced example")
    """

    _RECORD_TYPE = TextClassificationRecord

    def __init__(self, records: Optional[List[TextClassificationRecord]] = None):
        # we implement this to have more specific type hints
        super().__init__(records=records)

    @classmethod
    @requires_dependencies("datasets>1.17.0")
    def from_datasets(
        cls,
        dataset: "datasets.Dataset",
        text: Optional[str] = None,
        id: Optional[str] = None,
        inputs: Optional[Union[str, List[str]]] = None,
        annotation: Optional[str] = None,
        metadata: Optional[Union[str, List[str]]] = None,
    ) -> "DatasetForTextClassification":
        """Imports records from a `datasets.Dataset`.

        Columns that are not supported are ignored.

        Args:
            dataset: A datasets Dataset from which to import the records.
            text: The field name used as record text. Default: `None`
            id: The field name used as record id. Default: `None`
            inputs: A list of field names used for record inputs. Default: `None`
            annotation: The field name used as record annotation. Default: `None`
            metadata: The field name used as record metadata. Default: `None`

        Returns:
            The imported records in a argilla Dataset.

        Examples:
            >>> import datasets
            >>> ds = datasets.Dataset.from_dict({
            ...     "inputs": ["example"],
            ...     "prediction": [
            ...         [{"label": "LABEL1", "score": 0.9}, {"label": "LABEL2", "score": 0.1}]
            ...     ]
            ... })
            >>> DatasetForTextClassification.from_datasets(ds)
        """
        dataset, cols_to_be_joined = cls._prepare_dataset_and_column_mapping(
            dataset,
            dict(
                text=text,
                id=id,
                inputs=inputs,
                annotation=annotation,
                metadata=metadata,
            ),
        )

        records = []
        for row in dataset:
            row["inputs"] = cls._parse_inputs_field(row, cols_to_be_joined.get("inputs"))
            if row.get("inputs") is not None and row.get("text") is not None:
                del row["text"]

            if row.get("annotation") is not None:
                row["annotation"] = cls._parse_datasets_column_with_classlabel(
                    row["annotation"], dataset.features["annotation"]
                )

            if row.get("prediction"):
                row["prediction"] = (
                    [
                        (
                            pred["label"],
                            pred["score"],
                        )
                        for pred in row["prediction"]
                    ]
                    if row["prediction"] is not None
                    else None
                )

            if row.get("explanation"):
                row["explanation"] = (
                    {
                        key: [TokenAttributions(**tokattr_kwargs) for tokattr_kwargs in val]
                        for key, val in row["explanation"].items()
                    }
                    if row["explanation"] is not None
                    else None
                )

            if cols_to_be_joined.get("metadata"):
                row["metadata"] = cls._join_datasets_columns_and_delete(row, cols_to_be_joined["metadata"])

            records.append(TextClassificationRecord.parse_obj(row))

        return cls(records)

    @classmethod
    def _parse_inputs_field(
        cls,
        row: Dict[str, Any],
        columns: Optional[List[str]],
    ) -> Optional[Union[Dict[str, str], str]]:
        """Helper function to parse the inputs field.

        Args:
            row: A row of the dataset.Datasets
            columns: A list of columns to be joined for the inputs field, optional.

        Returns:
            None, a dictionary or a string as input for the inputs field.
        """
        inputs = row.get("inputs")

        if columns is not None:
            inputs = cls._join_datasets_columns_and_delete(row, columns)

        if isinstance(inputs, dict):
            inputs = {key: val for key, val in inputs.items() if val is not None}

        return inputs

    @classmethod
    def from_pandas(
        # we implement this to have more specific type hints
        cls,
        dataframe: pd.DataFrame,
    ) -> "DatasetForTextClassification":
        return super().from_pandas(dataframe)

    def _to_datasets_dict(self) -> Dict:
        # create a dict first, where we make the necessary transformations
        ds_dict = {}
        for key in self._RECORD_TYPE.__fields__:
            if key == "prediction":
                ds_dict[key] = [
                    [{"label": pred[0], "score": pred[1]} for pred in rec.prediction]
                    if rec.prediction is not None
                    else None
                    for rec in self._records
                ]
            elif key == "explanation":
                ds_dict[key] = [
                    {key: list(map(dict, tokattrs)) for key, tokattrs in rec.explanation.items()}
                    if rec.explanation is not None
                    else None
                    for rec in self._records
                ]
            elif key == "id":
                ds_dict[key] = [None if rec.id is None else str(rec.id) for rec in self._records]
            elif key == "metadata":
                ds_dict[key] = [getattr(rec, key) or None for rec in self._records]
            else:
                ds_dict[key] = [getattr(rec, key) for rec in self._records]

        return ds_dict

    @classmethod
    def _from_pandas(cls, dataframe: pd.DataFrame) -> "DatasetForTextClassification":
        return cls([TextClassificationRecord(**row) for row in dataframe.to_dict("records")])

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_transformers(
        self,
        train_size: Optional[float] = None,
        test_size: Optional[float] = None,
        seed: Optional[int] = None,
    ):
        import datasets

        ds_dict = {"id": [], "text": [], "label": []}
        for rec in self._records:
            if rec.annotation is None:
                continue

            if rec.text is not None:
                text = rec.text
            elif rec.text is None and "text" in rec.inputs:
                text = rec.inputs["text"]
            else:
                text = " ".join(rec.inputs.values())

            ds_dict["text"].append(text)
            ds_dict["label"].append(rec.annotation)
            ds_dict["id"].append(str(rec.id))

        if self._records[0].multi_label:
            labels = {label: None for labels in ds_dict["label"] for label in labels}
        else:
            labels = {label: None for label in ds_dict["label"]}

        class_label = (
            datasets.ClassLabel(names=sorted(labels.keys()))
            if ds_dict["label"]
            # in case we don't have any labels, ClassLabel fails with Dataset.from_dict({"labels": []})
            else datasets.Value("string")
        )

        feature_dict = {
            "id": datasets.Value("string"),
            "text": datasets.Value("string"),
            "label": [class_label] if self._records[0].multi_label else class_label,
        }

        ds = datasets.Dataset.from_dict(ds_dict, features=datasets.Features(feature_dict))

        if self._records[0].multi_label:
            require_dependencies("scikit-learn")
            from sklearn.preprocessing import MultiLabelBinarizer

            labels = [rec["label"] for rec in ds]
            mlb = MultiLabelBinarizer()
            binarized_labels = mlb.fit_transform(labels)
            feature_dict["binarized_label"] = feature_dict["label"]
            ds = datasets.Dataset.from_dict(
                {
                    "id": ds["id"],
                    "text": ds["text"],
                    "label": labels,
                    "binarized_label": binarized_labels,
                },
                features=datasets.Features(feature_dict),
            )
        if test_size is not None and test_size != 0:
            ds = ds.train_test_split(train_size=train_size, test_size=test_size, seed=seed)

        return ds

    @requires_dependencies("spacy")
    def _prepare_for_training_with_spacy(self, nlp: "spacy.Language", records: List[Record]) -> "spacy.tokens.DocBin":
        from spacy.tokens import DocBin

        db = DocBin(store_user_data=True)
        all_labels = self._verify_all_labels()

        # Creating the DocBin object as in https://spacy.io/usage/training#training-data

        for record in records:
            if record.annotation is None:
                continue

            if record.text is not None:
                text = record.text
            elif record.text is None and "text" in record.inputs:
                text = record.inputs["text"]
            else:
                text = ", ".join(f"{key}: {value}" for key, value in record.inputs.items())
            doc = nlp.make_doc(text)
            doc.user_data["id"] = record.id

            cats = dict.fromkeys(all_labels, 0)

            if isinstance(record.annotation, list):
                for anno in record.annotation:
                    cats[anno] = 1
            else:
                cats[record.annotation] = 1

            doc.cats = cats
            db.add(doc)

        return db

    def _prepare_for_training_with_spark_nlp(self, records: List[Record]) -> "pandas.DataFrame":
        if records[0].multi_label:
            label_name = "labels"
        else:
            label_name = "label"

        spark_nlp_data = []
        for record in records:
            if record.annotation is None:
                continue
            if record.id is None:
                record.id = str(uuid.uuid4())
            if record.text is not None:
                text = record.text
            elif record.text is None and "text" in record.inputs:
                text = record.inputs["text"]
            else:
                text = ", ".join(f"{key}: {value}" for key, value in record.inputs.items())

            spark_nlp_data.append([record.id, text, record.annotation])

        return pd.DataFrame(spark_nlp_data, columns=["id", "text", label_name])

    def _prepare_for_training_with_openai(self, **kwargs) -> "datasets.Dataset":
        """Prepares the dataset for training using the "openai" framework.

        Args:
            **kwargs: Specific to the task of the dataset.

        Returns:
            A pd.DataFrame.
        """
        separator = OPENAI_SEPARATOR
        whitespace = OPENAI_WHITESPACE
        self._verify_all_labels()  # verify that all labels are strings

        if len(self._records) <= len(self._SETTINGS.label_schema) * 100:
            warnings.warn("OpenAI recommends at least 100 examples per class for training a classification model.")

        jsonl = []
        for rec in self._records:
            if rec.annotation is None:
                continue

            if rec.text is not None:
                prompt = rec.text
            elif rec.text is None and "text" in rec.inputs:
                prompt = rec.inputs["text"]
            else:
                prompt = ", ".join(f"{key}: {value}" for key, value in rec.inputs.items())
            prompt += separator  # needed for better performance

            if self._records[0].multi_label:
                completion = " ".join([str(self._SETTINGS.label2id[annotation]) for annotation in rec.annotation])
            else:
                completion = str(self._SETTINGS.label2id[rec.annotation])

            jsonl.append({"id": rec.id, "prompt": prompt, "completion": whitespace + completion})

        return jsonl

    def _infer_settings_from_records(self) -> TextClassificationSettings:
        all_labels = set()
        for record in self._records:
            if record.annotation is None:
                continue
            elif isinstance(record.annotation, str):
                all_labels.add(record.annotation)
            elif isinstance(record.annotation, list):
                for label in record.annotation:
                    all_labels.add(label)
            else:
                # this is highly unlikely
                raise TypeError("Record.annotation contains an unsupported type: {}".format(type(record.annotation)))
        all_labels = list(all_labels)
        all_labels.sort()

        warnings.warn(
            f"No label schema provided. Using all_labels: TextClassificationSettings({all_labels}). "
            "We recommend providing a `TextClassificationSettings()` or setting "
            "`rg.configure_dataset_settings()`/`rg.load_dataset_settings()` to ensure reproducibility."
        )
        return TextClassificationSettings(all_labels)

    def _verify_all_labels(self):
        all_labels = self._SETTINGS.label_schema
        for record in self._records:
            if record.annotation is None:
                continue
            elif isinstance(record.annotation, str):
                if record.annotation not in all_labels:
                    raise ValueError(f"Label {record.annotation} is not in the settings.label_schema: {all_labels}.")
            elif isinstance(record.annotation, list):
                for label in record.annotation:
                    if label not in all_labels:
                        raise ValueError(f"Label {label} is not in the settings.label_schema: {all_labels}.")
            else:
                # this is highly unlikely
                raise TypeError("Record.annotation contains an unsupported type: {}".format(type(record.annotation)))

        return all_labels


@_prepend_docstring(TokenClassificationRecord)
class DatasetForTokenClassification(DatasetBase):
    """
    Examples:
        >>> # Import/export records:
        >>> import argilla as rg
        >>> dataset = rg.DatasetForTokenClassification.from_pandas(my_dataframe)
        >>> dataset.to_datasets()
        >>>
        >>> # Looping over the dataset:
        >>> assert len(dataset) == 2
        >>> for record in dataset:
        ...     print(record)
        >>>
        >>> # Passing in a list of records:
        >>> import argilla as rg
        >>> records = [
        ...     rg.TokenClassificationRecord(text="example", tokens=["example"]),
        ...     rg.TokenClassificationRecord(text="another example", tokens=["another", "example"]),
        ... ]
        >>> dataset = rg.DatasetForTokenClassification(records)
        >>>
        >>> # Indexing into the dataset:
        >>> dataset[0]
        ... rg.TokenClassificationRecord(text="example", tokens=["example"])
        >>> dataset[0] = rg.TokenClassificationRecord(text="replace example", tokens=["replace", "example"])
    """

    _RECORD_TYPE = TokenClassificationRecord

    def __init__(self, records: Optional[List[TokenClassificationRecord]] = None):
        # we implement this to have more specific type hints
        super().__init__(records=records)

    @classmethod
    def _record_init_args(cls) -> List[str]:
        """Adds the `tags` argument to default record init arguments"""
        parent_fields = super(DatasetForTokenClassification, cls)._record_init_args()
        return parent_fields + ["tags"]  # compute annotation from tags

    @classmethod
    @requires_dependencies("datasets>1.17.0")
    def from_datasets(
        cls,
        dataset: "datasets.Dataset",
        text: Optional[str] = None,
        id: Optional[str] = None,
        tokens: Optional[str] = None,
        tags: Optional[str] = None,
        metadata: Optional[Union[str, List[str]]] = None,
    ) -> "DatasetForTokenClassification":
        """Imports records from a `datasets.Dataset`.

        Columns that are not supported are ignored.

        Args:
            dataset: A datasets Dataset from which to import the records.
            text: The field name used as record text. Default: `None`
            id: The field name used as record id. Default: `None`
            tokens: The field name used as record tokens. Default: `None`
            tags: The field name used as record tags. Default: `None`
            metadata: The field name used as record metadata. Default: `None`

        Returns:
            The imported records in a argilla Dataset.

        Examples:
            >>> import datasets
            >>> ds = datasets.Dataset.from_dict({
            ...     "text": ["my example"],
            ...     "tokens": [["my", "example"]],
            ...     "prediction": [
            ...         [{"label": "LABEL1", "start": 3, "end": 10, "score": 1.0}]
            ...     ]
            ... })
            >>> DatasetForTokenClassification.from_datasets(ds)
        """
        dataset, cols_to_be_joined = cls._prepare_dataset_and_column_mapping(
            dataset,
            dict(
                text=text,
                tokens=tokens,
                tags=tags,
                id=id,
                metadata=metadata,
            ),
        )

        records = []
        for row in dataset:
            # TODO: fails with a KeyError if no tokens column is present and no mapping is indicated
            if not row["tokens"]:
                warnings.warn("Ignoring row with no tokens.")
                continue

            if row.get("tags"):
                row["tags"] = cls._parse_datasets_column_with_classlabel(row["tags"], dataset.features["tags"])

            if row.get("prediction"):
                row["prediction"] = cls.__entities_to_tuple__(row["prediction"])

            if row.get("annotation"):
                row["annotation"] = cls.__entities_to_tuple__(row["annotation"])

            if cols_to_be_joined.get("metadata"):
                row["metadata"] = cls._join_datasets_columns_and_delete(row, cols_to_be_joined["metadata"])

            records.append(TokenClassificationRecord.parse_obj(row))

        return cls(records)

    @classmethod
    def from_pandas(
        # we implement this to have more specific type hints
        cls,
        dataframe: pd.DataFrame,
    ) -> "DatasetForTokenClassification":
        return super().from_pandas(dataframe)

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_transformers(
        self,
        train_size: Optional[float] = None,
        test_size: Optional[float] = None,
        seed: Optional[int] = None,
    ):
        import datasets

        has_annotations = False
        for rec in self._records:
            if rec.annotation is not None:
                has_annotations = True
                break

        if not has_annotations:
            return datasets.Dataset.from_dict({})

        class_tags = ["O"]
        class_tags.extend([f"{pre}-{label}" for label in sorted(self._verify_all_labels()) for pre in ["B", "I"]])
        class_tags = datasets.ClassLabel(names=class_tags)

        def spans2iob(example):
            span_utils = SpanUtils(example["text"], example["tokens"])
            entity_spans = self.__entities_to_tuple__(example["annotation"])
            tags = span_utils.to_tags(entity_spans)

            return class_tags.str2int(tags)

        ds = self.to_datasets().filter(self.__only_annotations__).map(lambda example: {"ner_tags": spans2iob(example)})
        new_features = ds.features.copy()
        new_features["ner_tags"] = datasets.Sequence(feature=class_tags)
        ds = ds.cast(new_features)
        ds = ds.remove_columns(set(ds.column_names) - set(["id", "tokens", "ner_tags"]))

        if test_size is not None and test_size != 0:
            ds = ds.train_test_split(train_size=train_size, test_size=test_size, seed=seed)

        return ds

    @requires_dependencies("spacy")
    def _prepare_for_training_with_spacy(self, nlp: "spacy.Language", records: List[Record]) -> "spacy.tokens.DocBin":
        from spacy.tokens import DocBin

        db = DocBin(store_user_data=True)

        # Creating the DocBin object as in https://spacy.io/usage/training#training-data
        for record in records:
            if record.annotation is None:
                continue

            doc = nlp.make_doc(record.text)
            doc.user_data["id"] = record.id
            entities = []

            for anno in record.annotation:
                span = doc.char_span(anno[1], anno[2], label=anno[0])
                # There is a misalignment between record tokenization and spaCy tokenization
                if span is None:
                    # TODO(@dcfidalgo): Do we want to warn and continue or should we stop the training set generation?
                    raise ValueError(
                        "The following annotation does not align with the tokens"
                        " produced by the provided spacy language model:"
                        f" {(anno[0], record.text[anno[1]:anno[2]])}, {list(doc)}"
                    )
                else:
                    entities.append(span)

            doc.ents = entities
            db.add(doc)

        return db

    def _prepare_for_training_with_spark_nlp(self, records: List[Record]) -> "pandas.DataFrame":
        for record in records:
            if record.id is None:
                record.id = str(uuid.uuid4())
        iob_doc_data = [
            [
                record.id,
                record.text,
                record.tokens,
                [token["tag"] for token in record.metrics["tokens"]],
            ]
            for record in records
            if record.annotation is not None
        ]

        return pd.DataFrame(iob_doc_data, columns=["id", "text", "token", "label"])

    def _prepare_for_training_with_openai(self, **kwargs) -> "datasets.Dataset":
        """Prepares the dataset for training using the "openai" framework.

        Args:
            **kwargs: Specific to the task of the dataset.

        Returns:
            A pd.DataFrame.
        """
        separator = OPENAI_SEPARATOR
        end_token = OPENAI_END_TOKEN
        whitespace = OPENAI_WHITESPACE
        self._verify_all_labels()

        if len(self._records) <= 500:
            warnings.warn("OpenAI recommends at least 500 examples for training a conditional generation model.")

        jsonl = []
        for rec in self._records:
            if rec.annotation is None:
                continue

            prompt = rec.text + separator  # needed for better performance

            completion = {}
            for label, start, end in rec.annotation:
                completion[rec.text[start:end]] = self._SETTINGS.label2id[label]

            completion = "\n".join([f"- {k} {v}" for k, v in completion.items()])
            completion = completion + end_token
            jsonl.append({"id": rec.id, "prompt": prompt, "completion": whitespace + completion})

        return jsonl

    def _infer_settings_from_records(self) -> TokenClassificationSettings:
        all_labels = set()
        for record in self._records:
            if record.annotation:
                for label, _, _ in record.annotation:
                    all_labels.add(label)
        all_labels = list(all_labels)
        all_labels.sort()
        warnings.warn(
            f"No label schema provided. Using all_labels: TokenClassificationSettings({all_labels}). "
            "We recommend providing a `TokenClassificationSettings()` or setting "
            "`rg.configure_dataset_settings()`/`rg.load_dataset_settings()` to ensure reproducibility."
        )
        return TokenClassificationSettings(all_labels)

    def _verify_all_labels(self) -> List[str]:
        all_labels = self._SETTINGS.label_schema
        for record in self._records:
            if record.annotation:
                for label, _, _ in record.annotation:
                    if label not in all_labels:
                        raise ValueError(f"Label {label} is not in the settings.label_schema: {all_labels}.")

        return all_labels

    def __only_annotations__(self, data) -> bool:
        return data["annotation"] is not None

    def _to_datasets_dict(self) -> Dict:
        """Helper method to put token classification records in a `datasets.Dataset`"""

        # create a dict first, where we make the necessary transformations
        def entities_to_dict(
            entities: Optional[List[Union[Tuple[str, int, int, float], Tuple[str, int, int]]]]
        ) -> Optional[List[Dict[str, Union[str, int, float]]]]:
            if entities is None:
                return None
            return [
                {"label": ent[0], "start": ent[1], "end": ent[2]}
                if len(ent) == 3
                else {"label": ent[0], "start": ent[1], "end": ent[2], "score": ent[3]}
                for ent in entities
            ]

        ds_dict = {}
        for key in self._RECORD_TYPE.__fields__:
            if key == "prediction":
                ds_dict[key] = [entities_to_dict(rec.prediction) for rec in self._records]
            elif key == "annotation":
                ds_dict[key] = [entities_to_dict(rec.annotation) for rec in self._records]
            elif key == "id":
                ds_dict[key] = [None if rec.id is None else str(rec.id) for rec in self._records]
            elif key == "metadata":
                ds_dict[key] = [getattr(rec, key) or None for rec in self._records]
            else:
                ds_dict[key] = [getattr(rec, key) for rec in self._records]

        return ds_dict

    @staticmethod
    def __entities_to_tuple__(
        entities,
    ) -> List[Union[Tuple[str, int, int], Tuple[str, int, int, float]]]:
        return [
            (ent["label"], ent["start"], ent["end"])
            if len(ent) == 3
            else (ent["label"], ent["start"], ent["end"], ent["score"] or 0.0)
            for ent in entities
        ]

    @classmethod
    def _from_pandas(cls, dataframe: pd.DataFrame) -> "DatasetForTokenClassification":
        return cls([TokenClassificationRecord(**row) for row in dataframe.to_dict("records")])


@_prepend_docstring(Text2TextRecord)
class DatasetForText2Text(DatasetBase):
    """
    Examples:
        >>> # Import/export records:
        >>> import argilla as rg
        >>> dataset = rg.DatasetForText2Text.from_pandas(my_dataframe)
        >>> dataset.to_datasets()
        >>>
        >>> # Passing in a list of records:
        >>> records = [
        ...     rg.Text2TextRecord(text="example"),
        ...     rg.Text2TextRecord(text="another example"),
        ... ]
        >>> dataset = rg.DatasetForText2Text(records)
        >>> assert len(dataset) == 2
        >>>
        >>> # Looping over the dataset:
        >>> for record in dataset:
        ...     print(record)
        >>>
        >>> # Indexing into the dataset:
        >>> dataset[0]
        ... rg.Text2TextRecord(text="example"})
        >>> dataset[0] = rg.Text2TextRecord(text="replaced example")
    """

    _RECORD_TYPE = Text2TextRecord

    def __init__(self, records: Optional[List[Text2TextRecord]] = None):
        # we implement this to have more specific type hints
        super().__init__(records=records)

    @classmethod
    @requires_dependencies("datasets>1.17.0")
    def from_datasets(
        cls,
        dataset: "datasets.Dataset",
        text: Optional[str] = None,
        annotation: Optional[str] = None,
        metadata: Optional[Union[str, List[str]]] = None,
        id: Optional[str] = None,
    ) -> "DatasetForText2Text":
        """Imports records from a `datasets.Dataset`.

        Columns that are not supported are ignored.

        Args:
            dataset: A datasets Dataset from which to import the records.
            text: The field name used as record text. Default: `None`
            annotation: The field name used as record annotation. Default: `None`
            metadata: The field name used as record metadata. Default: `None`

        Returns:
            The imported records in a argilla Dataset.

        Examples:
            >>> import datasets
            >>> ds = datasets.Dataset.from_dict({
            ...     "text": ["my example"],
            ...     "prediction": [["mi ejemplo", "ejemplo mio"]]
            ... })
            >>> # or
            >>> ds = datasets.Dataset.from_dict({
            ...     "text": ["my example"],
            ...     "prediction": [[{"text": "mi ejemplo", "score": 0.9}]]
            ... })
            >>> DatasetForText2Text.from_datasets(ds)
        """
        dataset, cols_to_be_joined = cls._prepare_dataset_and_column_mapping(
            dataset,
            dict(
                text=text,
                annotation=annotation,
                id=id,
                metadata=metadata,
            ),
        )

        records = []
        for row in dataset:
            if row.get("prediction"):
                row["prediction"] = cls._parse_prediction_field(row["prediction"])

            if cols_to_be_joined.get("metadata"):
                row["metadata"] = cls._join_datasets_columns_and_delete(row, cols_to_be_joined["metadata"])

            records.append(Text2TextRecord.parse_obj(row))

        return cls(records)

    @staticmethod
    def _parse_prediction_field(predictions: List[Union[str, Dict[str, str]]]):
        def extract_prediction(prediction: Union[str, Dict]):
            if isinstance(prediction, str):
                return prediction
            if prediction["score"] is None:
                return prediction["text"]
            return prediction["text"], prediction["score"]

        return [extract_prediction(pred) for pred in predictions]

    @classmethod
    def from_pandas(
        # we implement this to have more specific type hints
        cls,
        dataframe: pd.DataFrame,
    ) -> "DatasetForText2Text":
        return super().from_pandas(dataframe)

    def _to_datasets_dict(self) -> Dict:
        # create a dict first, where we make the necessary transformations
        def pred_to_dict(pred: Union[str, Tuple[str, float]]):
            if isinstance(pred, str):
                return {"text": pred, "score": None}
            return {"text": pred[0], "score": pred[1]}

        ds_dict = {}
        for key in self._RECORD_TYPE.__fields__:
            if key == "prediction":
                ds_dict[key] = [
                    [pred_to_dict(pred) for pred in rec.prediction] if rec.prediction is not None else None
                    for rec in self._records
                ]
            elif key == "id":
                ds_dict[key] = [None if rec.id is None else str(rec.id) for rec in self._records]
            elif key == "metadata":
                ds_dict[key] = [getattr(rec, key) or None for rec in self._records]
            else:
                ds_dict[key] = [getattr(rec, key) for rec in self._records]

        return ds_dict

    @classmethod
    def _from_pandas(cls, dataframe: pd.DataFrame) -> "DatasetForText2Text":
        return cls([Text2TextRecord(**row) for row in dataframe.to_dict("records")])

    @requires_dependencies("datasets>1.17.0")
    def _prepare_for_training_with_transformers(
        self,
        train_size: Optional[float] = None,
        test_size: Optional[float] = None,
        seed: Optional[int] = None,
    ):
        import datasets

        ds_dict = {"id": [], "text": [], "target": []}
        for rec in self._records:
            if rec.annotation is None:
                continue
            ds_dict["id"].append(rec.id)
            ds_dict["text"].append(rec.text)
            ds_dict["target"].append(rec.annotation)

        feature_dict = {
            "id": datasets.Value("string"),
            "text": datasets.Value("string"),
            "target": datasets.Value("string"),
        }

        ds = datasets.Dataset.from_dict(ds_dict, features=datasets.Features(feature_dict))

        if test_size is not None and test_size != 0:
            ds = ds.train_test_split(train_size=train_size, test_size=test_size, seed=seed)

        return ds

    def _prepare_for_training_with_spark_nlp(self, records: List[Record]) -> "pandas.DataFrame":
        spark_nlp_data = []
        for record in records:
            if record.annotation is None:
                continue
            if record.id is None:
                record.id = str(uuid.uuid4())
            text = record.text

            spark_nlp_data.append([record.id, text, record.annotation])

        return pd.DataFrame(spark_nlp_data, columns=["id", "text", "target"])

    def _prepare_for_training_with_openai(self, **kwargs) -> "datasets.Dataset":
        """Prepares the dataset for training using the "openai" framework.

        Args:
            **kwargs: Specific to the task of the dataset.

        Returns:
            A pd.DataFrame.
        """
        separator = OPENAI_SEPARATOR
        end_token = OPENAI_END_TOKEN
        whitespace = OPENAI_WHITESPACE
        if len(self._records) <= 500:
            warnings.warn("OpenAI recommends at least 500 examples for training a conditional generation model.")

        jsonl = []
        for rec in self._records:
            if rec.annotation is None:
                continue

            prompt = rec.text + separator  # needed for better performance
            completion = rec.annotation + end_token

            jsonl.append({"id": rec.id, "prompt": prompt, "completion": whitespace + completion})

        return jsonl


Dataset = Union[DatasetForTextClassification, DatasetForTokenClassification, DatasetForText2Text]


def read_datasets(dataset: "datasets.Dataset", task: Union[str, TaskType], **kwargs) -> Dataset:
    """Reads a datasets Dataset and returns a argilla Dataset

    Columns not supported by the :mod:`Record <argilla.client.models>` instance corresponding
    with the task are ignored.

    Args:
        dataset: Dataset to be read in.
        task: Task for the dataset, one of: ["TextClassification", "TokenClassification", "Text2Text"].
        **kwargs: Passed on to the task-specific ``DatasetFor*.from_datasets()`` method.

    Returns:
        A argilla dataset for the given task.

    Examples:
        >>> # Read text classification records from a datasets Dataset
        >>> import datasets
        >>> ds = datasets.Dataset.from_dict({
        ...     "inputs": ["example"],
        ...     "prediction": [
        ...         [{"label": "LABEL1", "score": 0.9}, {"label": "LABEL2", "score": 0.1}]
        ...     ]
        ... })
        >>> read_datasets(ds, task="TextClassification")
        >>>
        >>> # Read token classification records from a datasets Dataset
        >>> ds = datasets.Dataset.from_dict({
        ...     "text": ["my example"],
        ...     "tokens": [["my", "example"]],
        ...     "prediction": [
        ...         [{"label": "LABEL1", "start": 3, "end": 10}]
        ...     ]
        ... })
        >>> read_datasets(ds, task="TokenClassification")
        >>>
        >>> # Read text2text records from a datasets Dataset
        >>> ds = datasets.Dataset.from_dict({
        ...     "text": ["my example"],
        ...     "prediction": [["mi ejemplo", "ejemplo mio"]]
        ... })
        >>> # or
        >>> ds = datasets.Dataset.from_dict({
        ...     "text": ["my example"],
        ...     "prediction": [[{"text": "mi ejemplo", "score": 0.9}]]
        ... })
        >>> read_datasets(ds, task="Text2Text")
    """
    if isinstance(task, str):
        task = TaskType(task)

    if task is TaskType.text_classification:
        return DatasetForTextClassification.from_datasets(dataset, **kwargs)
    if task is TaskType.token_classification:
        return DatasetForTokenClassification.from_datasets(dataset, **kwargs)
    if task is TaskType.text2text:
        return DatasetForText2Text.from_datasets(dataset, **kwargs)

    raise NotImplementedError("Reading a datasets Dataset is not implemented for the given task!")


def read_pandas(dataframe: pd.DataFrame, task: Union[str, TaskType]) -> Dataset:
    """Reads a pandas DataFrame and returns a argilla Dataset

    Columns not supported by the :mod:`Record <argilla.client.models>` instance corresponding
    with the task are ignored.

    Args:
        dataframe: Dataframe to be read in.
        task: Task for the dataset, one of: ["TextClassification", "TokenClassification", "Text2Text"]

    Returns:
        A argilla dataset for the given task.

    Examples:
        >>> # Read text classification records from a pandas DataFrame
        >>> import pandas as pd
        >>> df = pd.DataFrame({
        ...     "inputs": ["example"],
        ...     "prediction": [
        ...         [("LABEL1", 0.9), ("LABEL2", 0.1)]
        ...     ]
        ... })
        >>> read_pandas(df, task="TextClassification")
        >>>
        >>> # Read token classification records from a datasets Dataset
        >>> df = pd.DataFrame({
        ...     "text": ["my example"],
        ...     "tokens": [["my", "example"]],
        ...     "prediction": [
        ...         [("LABEL1", 3, 10)]
        ...     ]
        ... })
        >>> read_pandas(df, task="TokenClassification")
        >>>
        >>> # Read text2text records from a datasets Dataset
        >>> df = pd.DataFrame({
        ...     "text": ["my example"],
        ...     "prediction": [["mi ejemplo", "ejemplo mio"]]
        ... })
        >>> # or
        >>> ds = pd.DataFrame({
        ...     "text": ["my example"],
        ...     "prediction": [[("mi ejemplo", 0.9)]]
        ... })
        >>> read_pandas(df, task="Text2Text")
    """
    if isinstance(task, str):
        task = TaskType(task)

    if task is TaskType.text_classification:
        return DatasetForTextClassification.from_pandas(dataframe)
    if task is TaskType.token_classification:
        return DatasetForTokenClassification.from_pandas(dataframe)
    if task is TaskType.text2text:
        return DatasetForText2Text.from_pandas(dataframe)
    raise NotImplementedError("Reading a pandas DataFrame is not implemented for the given task!")


class WrongRecordTypeError(Exception):
    pass



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

import typer

from argilla.cli.callback import init_callback
from argilla.cli.datasets.delete import delete_dataset
from argilla.cli.datasets.list import list_datasets
from argilla.cli.datasets.push import push_to_huggingface

_COMMANDS_REQUIRING_DATASET = ["delete", "push-to-huggingface"]


def callback(
    ctx: typer.Context,
    name: Optional[str] = typer.Option(None, help="The name of the `FeedbackDataset` to which apply the command"),
    workspace: Optional[str] = typer.Option(None, help="The name of the workspace where the `FeedbackDataset` belongs"),
) -> None:
    init_callback()

    from argilla.cli.rich import echo_in_panel
    from argilla.client.feedback.dataset.local.dataset import FeedbackDataset

    if ctx.invoked_subcommand not in _COMMANDS_REQUIRING_DATASET:
        return

    if name is None:
        raise typer.BadParameter("The command requires a workspace name provided using '--name' option")

    try:
        dataset = FeedbackDataset.from_argilla(name=name, workspace=workspace)
    except ValueError as e:
        echo_in_panel(
            f"`FeedbackDataset` with name={name} not found in Argilla. Try using '--workspace' option."
            if not workspace
            else f"`FeedbackDataset with name={name} and workspace={workspace} not found in Argilla.",
            title="Dataset not found",
            title_align="left",
            success=False,
        )
        raise typer.Exit(1) from e
    except Exception as e:
        echo_in_panel(
            "An unexpected error occurred when trying to get the `FeedbackDataset` from Argilla.",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e

    ctx.obj = dataset


app = typer.Typer(help="Commands for dataset management", no_args_is_help=True, callback=callback)

app.command(name="list", help="List datasets linked to user's workspaces")(list_datasets)
app.command(name="push-to-huggingface", help="Push a dataset to HuggingFace Hub")(push_to_huggingface)
app.command(name="delete", help="Deletes a dataset")(delete_dataset)


if __name__ == "__main__":
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer

from argilla.cli.rich import echo_in_panel
from argilla.client.users import User


def get_user(username: str) -> User:
    try:
        return User.from_name(username)
    except ValueError as e:
        echo_in_panel(
            f"User with username={username} does not exist", title="User not found", title_align="left", success=False
        )
        raise typer.Exit(code=1) from e
    except RuntimeError as e:
        echo_in_panel(
            "An unexpected error occurred when trying to retrieve the user from the Argilla server",
            title="Unexpected error",
            title_align="left",
            success=False,
        )
        raise typer.Exit(code=1) from e



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

import httpx

from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import ErrorMessage, HTTPValidationError, Response
from argilla.client.sdk.v1.datasets.models import FeedbackSuggestionModel
from argilla.client.sdk.v1.records.models import FeedbackItemModel


def update_record(
    # TODO: Use the proper sdk API Model instead of the dict
    client: httpx.Client,
    id: Union[str, UUID],
    data: Dict[str, Any],
) -> Response[Union[FeedbackItemModel, ErrorMessage, HTTPValidationError]]:
    """Sends a `PATCH` request to `/api/v1/records/{id}` endpoint to partially update
    a record in Argilla.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the record to be updated in Argilla.
        data: the data to be updated in the record. It can contain `metadata` and/or
            `suggestions` fields.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a `FeedbackItemModel`.
    """
    url = f"/api/v1/records/{id}"

    body = {}
    if "metadata" in data:
        body["metadata"] = data["metadata"]
    if "suggestions" in data:
        body["suggestions"] = data["suggestions"]
    if "vectors" in data:
        body["vectors"] = data["vectors"]

    response = client.patch(url=url, json=body)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackItemModel.parse_raw(response.content)
        return response_obj

    return handle_response_error(response)


def delete_record(
    client: httpx.Client, id: UUID
) -> Response[Union[FeedbackItemModel, ErrorMessage, HTTPValidationError]]:
    """Sends a DELETE request to `/api/v1/records/{id}` endpoint to delete a
    record from Argilla.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the record to be deleted in Argilla.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a `FeedbackItemModel`.
    """
    url = f"/api/v1/records/{id}"

    response = client.delete(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackItemModel.parse_raw(response.content)
        return response_obj
    return handle_response_error(response)


def delete_suggestions(
    client: httpx.Client, id: UUID, suggestion_ids: List[UUID]
) -> Response[Union[ErrorMessage, HTTPValidationError]]:
    """Sends a DELETE request to `/api/v1/records/{id}/suggestions` endpoint to delete
    suggestions from a record in Argilla.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the record in Argilla.
        suggestion_ids: the ids of the suggestions to be deleted from the record.

    Returns:
        A `Response` object with the response itself, and/or the error codes if applicable.
    """
    url = f"/api/v1/records/{id}/suggestions"
    params = {"ids": ",".join([str(suggestion_id) for suggestion_id in suggestion_ids])}

    response = client.delete(url=url, params=params)

    if response.status_code == 204:
        return Response.from_httpx_response(response)
    return handle_response_error(response)


def set_suggestion(
    client: httpx.Client,
    record_id: UUID,
    question_id: UUID,
    value: Any,
    type: Optional[Literal["model", "human"]] = None,
    score: Optional[float] = None,
    agent: Optional[str] = None,
) -> Response[Union[FeedbackSuggestionModel, ErrorMessage, HTTPValidationError]]:
    """Sends a PUT request to `/api/v1/records/{id}/suggestions` endpoint to add or update
    a suggestion for a question in the `FeedbackDataset`.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        record_id: the id of the record to add the suggestion to.
        question_id: the id of the question to add the suggestion to.
        value: the value of the suggestion.
        type: the type of the suggestion. It can be either `model` or `human`. Defaults to None.
        score: the score of the suggestion. Defaults to None.
        agent: the agent used to obtain the suggestion. Defaults to None.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if the
        request was successful, which is a `FeedbackSuggestionModel`.
    """
    url = f"/api/v1/records/{record_id}/suggestions"

    suggestion = {
        "question_id": str(question_id),
        "value": value,
    }
    if type is not None:
        suggestion["type"] = type
    if score is not None:
        suggestion["score"] = score
    if agent is not None:
        suggestion["agent"] = agent

    response = client.put(url=url, json=suggestion)

    if response.status_code in [200, 201]:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackSuggestionModel(**response.json())
        return response_obj
    return handle_response_error(response)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional, Union
from uuid import UUID

import httpx

from argilla.client.sdk.commons.errors_handler import handle_response_error
from argilla.client.sdk.commons.models import ErrorMessage, HTTPValidationError, Response
from argilla.client.sdk.v1.metadata_properties.models import FeedbackMetadataPropertyModel


def delete_metadata_property(
    client: httpx.Client, id: UUID
) -> Response[Union[FeedbackMetadataPropertyModel, ErrorMessage, HTTPValidationError]]:
    """Sends a DELETE request to `/api/v1/metadata_properties/{id}` endpoint to delete a
    metadata property from a `FeedbackTask` dataset in Argilla.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the metadata property to be deleted in Argilla.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a `FeedbackMetadataPropertyModel`.
    """
    url = f"/api/v1/metadata-properties/{id}"

    response = client.delete(url=url)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackMetadataPropertyModel.parse_raw(response.content)
        return response_obj
    return handle_response_error(response)


def update_metadata_property(
    client: httpx.Client,
    id: UUID,
    title: Optional[str] = None,
    visible_for_annotators: Optional[bool] = None,
) -> Response[Union[FeedbackMetadataPropertyModel, ErrorMessage, HTTPValidationError]]:
    """Sends a PATCH request to `/api/v1/metadata_properties/{id}` endpoint to update a
    metadata property from a `FeedbackTask` dataset in Argilla.

    Args:
        client: the authenticated Argilla client to be used to send the request to the API.
        id: the id of the metadata property to be updated in Argilla.
        title: the new title of the metadata property. Defaults to `None`.
        visible_for_annotators: whether the metadata property should be visible for annotators. Defaults to `None`.

    Returns:
        A `Response` object containing a `parsed` attribute with the parsed response if
        the request was successful, which is a `FeedbackMetadataPropertyModel`.
    """
    url = f"/api/v1/metadata-properties/{id}"

    payload = {}
    if title is not None:
        payload["title"] = title
    if visible_for_annotators is not None:
        payload["visible_for_annotators"] = visible_for_annotators

    response = client.patch(url=url, json=payload)

    if response.status_code == 200:
        response_obj = Response.from_httpx_response(response)
        response_obj.parsed = FeedbackMetadataPropertyModel.parse_raw(response.content)
        return response_obj
    return handle_response_error(response)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import math
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

from argilla.client.feedback.schemas.enums import MetadataPropertyTypes
from argilla.client.feedback.schemas.validators import (
    validate_numeric_metadata_filter_bounds,
    validate_numeric_metadata_property_bounds,
)
from argilla.pydantic_v1 import (
    BaseModel,
    Extra,
    Field,
    StrictFloat,
    StrictInt,
    StrictStr,
    ValidationError,
    root_validator,
    validator,
)

TERMS_METADATA_PROPERTY_MIN_VALUES = 1
TERMS_METADATA_FILTER_MIN_VALUES = 1


class MetadataPropertySchema(BaseModel, ABC):
    """Base schema for the `FeedbackDataset` metadata properties.

    Args:
        name: The name of the metadata property.
        title: A title of the metadata property (what's shown in the UI). Defaults to `None`,
            which means that if not provided then the `name` will be used as the `title`.
        visible_for_annotators: Whether the metadata property should be visible for
            users with the `annotator` role. Defaults to `True`.
        type: The type of the metadata property. A value should be set for this
            attribute in the class inheriting from this one to be able to use a
            discriminated union based on the `type` field. Defaults to `None`.

    Disclaimer:
        You should not use this class directly, but instead use the classes that inherit
        from this one, as they will have the `type` field already defined, and ensured
        to be supported by Argilla.
    """

    name: str = Field(..., regex=r"^(?=.*[a-z0-9])[a-z0-9_-]+$")
    title: Optional[str] = None
    visible_for_annotators: Optional[bool] = True
    type: MetadataPropertyTypes = Field(..., allow_mutation=False)

    class Config:
        validate_assignment = True
        extra = Extra.forbid
        exclude = {"type"}

    @validator("title", always=True)
    def title_must_have_value(cls, v: Optional[str], values: Dict[str, Any]) -> str:
        return values.get("name") if not v else v

    @property
    @abstractmethod
    def server_settings(self) -> Dict[str, Any]:
        ...

    def to_server_payload(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "title": self.title,
            "visible_for_annotators": self.visible_for_annotators,
            "settings": self.server_settings,
        }

    @property
    @abstractmethod
    def _pydantic_field_with_validator(self) -> Tuple[Dict[str, Tuple[Any, ...]], Dict[str, Callable]]:
        ...

    @abstractmethod
    def _validate_filter(self, metadata_filter: "MetadataFilters") -> None:
        pass

    @abstractmethod
    def _check_allowed_value_type(self, value: Any) -> Any:
        ...

    @abstractmethod
    def _validator(self, value: Any) -> Any:
        ...


def _validator_definition(schema: MetadataPropertySchema) -> Dict[str, Any]:
    """Returns a dictionary with the pydantic validator definition for the provided schema."""
    return {f"{schema.name}_validator": validator(schema.name, allow_reuse=True, pre=True)(schema._validator)}


class TermsMetadataProperty(MetadataPropertySchema):
    """Schema for the `FeedbackDataset` metadata properties of type `terms`. This kind
    of metadata property will be used for filtering the metadata of a record based on
    a list of possible terms or values.

    Args:
        name: The name of the metadata property.
        title: A title of the metadata property (what's shown in the UI). Defaults to `None`,
            which means that if not provided then the `name` will be used as the `title`.
        visible_for_annotators: Whether the metadata property should be visible for
            users with the `annotator` role. Defaults to `True`.
        values: A list of possible values for the metadata property. It must contain
            at least one value.

    Examples:
        >>> from argilla.client.feedback.schemas.metadata import TermsMetadataProperty
        >>> TermsMetadataProperty(name="color", values=["red", "blue", "green"])
    """

    type: Literal[MetadataPropertyTypes.terms] = Field(MetadataPropertyTypes.terms.value, allow_mutation=False)
    values: Optional[List[str]] = None

    @validator("values")
    def check_values(cls, terms_values: Union[List[str], None], values: Dict[str, Any]) -> List[str]:
        if terms_values is not None:
            if len(terms_values) < TERMS_METADATA_PROPERTY_MIN_VALUES:
                raise ValueError(
                    f"`TermsMetadataProperty` with name={values.get('name')} must have at least {TERMS_METADATA_PROPERTY_MIN_VALUES} `values`"
                )
            if len(set(terms_values)) != len(terms_values):
                raise ValueError(
                    f"`TermsMetadataProperty` with name={values.get('name')} cannot have repeated `values`"
                )
        return terms_values

    @property
    def server_settings(self) -> Dict[str, Any]:
        settings: Dict[str, Any] = {"type": self.type}
        if self.values is not None:
            settings["values"] = self.values
        return settings

    def _all_values_exist(self, introduced_value: Optional[Union[str, List[str]]] = None) -> Optional[str]:
        if introduced_value is None or self.values is None:
            return introduced_value

        if isinstance(introduced_value, str):
            values = [introduced_value]
        else:
            values = introduced_value

        for value in values:
            if value not in self.values:
                raise ValueError(
                    f"Provided '{self.name}={value}' is not valid, only values in {self.values} are allowed."
                )

        return introduced_value

    def _check_allowed_value_type(self, value: Any) -> Any:
        if value is None or isinstance(value, str):
            return value

        if isinstance(value, list):
            return [self._check_allowed_value_type(v) for v in value]

        raise TypeError(
            f"Provided '{self.name}={value}' of type {type(value)} is not valid, "
            "only values of type `str` or `list` are allowed."
        )

    def _validator(self, value: Any) -> Any:
        return self._all_values_exist(self._check_allowed_value_type(value))

    @property
    def _pydantic_field_with_validator(self) -> Tuple[Dict[str, tuple], Dict[str, Callable]]:
        # TODO: Simplify the validation logic and do not base on dynamic pydantic models
        field_type, default_value = Optional[Union[StrictStr, List[StrictStr]]], None

        return {self.name: (field_type, default_value)}, _validator_definition(self)

    def _validate_filter(self, metadata_filter: "TermsMetadataFilter") -> None:
        if self.values is not None and not all(value in self.values for value in metadata_filter.values):
            raise ValidationError(
                f"Provided 'values={metadata_filter.values}' is not valid, only values in {self.values} are allowed.",
                model=type(metadata_filter),
            )


class _NumericMetadataPropertySchema(MetadataPropertySchema):
    """Protected schema for the numeric `FeedbackDataset` metadata properties.

    Args:
        name: The name of the metadata property.
        title: A title of the metadata property (what's shown in the UI). Defaults to `None`,
            which means that if not provided then the `name` will be used as the `title`.
        visible_for_annotators: Whether the metadata property should be visible for
            users with the `annotator` role. Defaults to `True`.
        min: The lower bound of the numeric value. Must be provided and be lower than
            the `max` value.
        max: The upper bound of the numeric value. Must be provided and be greater
            than the `min` value.

    Disclaimer:
        You should not use this class directly, but instead use the classes that inherit
        from this one.
    """

    min: Optional[Union[int, float]] = None
    max: Optional[Union[int, float]] = None

    _bounds_validator = root_validator(allow_reuse=True, skip_on_failure=True)(
        validate_numeric_metadata_property_bounds
    )

    @property
    def server_settings(self) -> Dict[str, Any]:
        settings: Dict[str, Any] = {"type": self.type}
        if self.min is not None:
            settings["min"] = self.min
        if self.max is not None:
            settings["max"] = self.max
        return settings

    def _value_in_bounds(self, provided_value: Optional[Union[int, float]]) -> Union[int, float]:
        if provided_value is not None:
            if (self.min is not None and self.min > provided_value) or (
                self.max is not None and self.max < provided_value
            ):
                if self.min is not None and self.max is not None:
                    raise ValueError(
                        f"Provided '{self.name}={provided_value}' is not valid, only values between {self.min} and {self.max} are allowed."
                    )
                if self.min is not None:
                    raise ValueError(
                        f"Provided '{self.name}={provided_value}' is not valid, only values over {self.min} are allowed."
                    )
                if self.max is not None:
                    raise ValueError(
                        f"Provided '{self.name}={provided_value}' is not valid, only values under {self.max} are allowed."
                    )
        return provided_value

    def _check_nan(self, provided_value: Optional[Union[int, float]]) -> Optional[Union[int, float]]:
        if provided_value != provided_value:
            raise ValueError(f"Provided '{self.name}={provided_value}' is not valid, NaN values are not allowed.")

        return provided_value

    def _validator(self, value: Any) -> Any:
        return self._value_in_bounds(self._check_allowed_value_type(self._check_nan(value)))

    def _validate_filter(self, metadata_filter: Union["IntegerMetadataFilter", "FloatMetadataFilter"]) -> None:
        metadata_filter_ = metadata_filter.dict()
        for allowed_arg in ["ge", "le"]:
            if metadata_filter_[allowed_arg] is not None:
                if (
                    self.max is not None
                    and self.min is not None
                    and not (self.max >= metadata_filter_[allowed_arg] >= self.min)
                ):
                    raise ValidationError(
                        f"Provided '{allowed_arg}={metadata_filter_[allowed_arg]}' is not valid, "
                        f"only values between {self.min} and {self.max} are allowed.",
                        model=type(metadata_filter),
                    )
                if self.max is not None and not (self.max >= metadata_filter_[allowed_arg]):
                    raise ValidationError(
                        f"Provided '{allowed_arg}={metadata_filter_[allowed_arg]}' is not valid, "
                        f"only values under {self.max} are allowed.",
                        model=type(metadata_filter),
                    )
                if self.min is not None and not (self.min <= metadata_filter_[allowed_arg]):
                    raise ValidationError(
                        f"Provided '{allowed_arg}={metadata_filter_[allowed_arg]}' is not valid, "
                        f"only values over {self.min} are allowed.",
                        model=type(metadata_filter),
                    )


class IntegerMetadataProperty(_NumericMetadataPropertySchema):
    """Schema for the `FeedbackDataset` metadata properties of type `integer`. This kind
    of metadata property will be used for filtering the metadata of a record based on
    an integer value to which `min` and `max` filters can be applied.

    Args:
        name: The name of the metadata property.
        title: A title of the metadata property (what's shown in the UI). Defaults to `None`,
            which means that if not provided then the `name` will be used as the `title`.
        visible_for_annotators: Whether the metadata property should be visible for
            users with the `annotator` role. Defaults to `True`.
        min: The lower bound of the integer value. Must be provided, and be lower than
            the `max` value.
        max: The upper bound of the integer value. Must be provided, and be greater than
            the `min` value.

    Examples:
        >>> from argilla.client.feedback.schemas.metadata import IntegerMetadataProperty
        >>> IntegerMetadataProperty(name="day", min=0, max=31)
    """

    type: Literal[MetadataPropertyTypes.integer] = Field(MetadataPropertyTypes.integer.value, allow_mutation=False)
    min: Optional[int] = None
    max: Optional[int] = None

    @property
    def _pydantic_field_with_validator(
        self,
    ) -> Tuple[Dict[str, tuple], Dict[str, Callable]]:
        field_type, default_value = Optional[StrictInt], None
        return {self.name: (field_type, default_value)}, _validator_definition(self)

    def _check_allowed_value_type(self, value: Any) -> Any:
        if value is not None:
            if isinstance(value, int):
                return value
            raise TypeError(
                f"Provided '{self.name}={value}' of type {type(value)} is not valid, "
                "only values of type `int` are allowed."
            )
        return value


class FloatMetadataProperty(_NumericMetadataPropertySchema):
    """Schema for the `FeedbackDataset` metadata properties of type `float`. This kind
    of metadata property will be used for filtering the metadata of a record based on
    an float value to which `min` and `max` filters can be applied.

    Args:
        name: The name of the metadata property.
        title: A title of the metadata property (what's shown in the UI). Defaults to `None`,
            which means that if not provided then the `name` will be used as the `title`.
        visible_for_annotators: Whether the metadata property should be visible for
            users with the `annotator` role. Defaults to `True`.
        min: The lower bound of the float value. Must be provided, and be lower than
            the `max` value.
        max: The upper bound of the float value. Must be provided, and be greater than
            the `min` value.

    Examples:
        >>> from argilla.client.feedback.schemas.metadata import FloatMetadataProperty
        >>> FloatMetadataProperty(name="price", min=0.0, max=100.0)
    """

    type: Literal[MetadataPropertyTypes.float] = Field(MetadataPropertyTypes.float.value, allow_mutation=False)
    min: Optional[float] = None
    max: Optional[float] = None

    def _check_allowed_value_type(self, value: Any) -> Any:
        if value is None or isinstance(value, (int, float)):
            return value

        raise TypeError(
            f"Provided '{self.name}={value}' of type {type(value)} is not valid, "
            "only values of type `int` or `float` are allowed."
        )

    @property
    def _pydantic_field_with_validator(
        self,
    ) -> Tuple[Dict[str, tuple], Dict[str, Any]]:
        field_type, default_value = Optional[StrictFloat], None
        return {self.name: (field_type, default_value)}, _validator_definition(self)


class MetadataFilterSchema(BaseModel, ABC):
    """Base schema for the `FeedbackDataset` metadata filters.

    Args:
        name: The name of the metadata property.
        type: The type of the metadata property. A value should be set for this
            attribute in the class inheriting from this one to be able to use a
            discriminated union based on the `type` field.

    Disclaimer:
        You should not use this class directly, but instead use the classes that inherit
        from this one, as they will have the `type` field already defined, and ensured
        to be supported by Argilla.
    """

    name: str = Field(..., regex=r"^(?=.*[a-z0-9])[a-z0-9_-]+$")
    type: MetadataPropertyTypes = Field(..., allow_mutation=False)

    class Config:
        validate_assignment = True
        extra = Extra.forbid
        exclude = {"type"}

    @property
    @abstractmethod
    def query_string(self) -> str:
        ...


class TermsMetadataFilter(MetadataFilterSchema):
    """Schema for the `FeedbackDataset` metadata filters of type `terms`. This kind
    of metadata filter will be used for filtering the metadata of a record based on
    a list of possible terms or values.

    Args:
        name: The name of the metadata property.
        values: A list of possible values for the metadata property. It must contain
            at least two values.

    Examples:
        >>> from argilla.client.feedback.schemas.metadata import TermsMetadataFilter
        >>> TermsMetadataFilter(name="color", values=["red", "blue", "green"])
    """

    type: Literal[MetadataPropertyTypes.terms] = Field(MetadataPropertyTypes.terms.value, allow_mutation=False)
    values: List[str] = Field(..., min_items=TERMS_METADATA_FILTER_MIN_VALUES)

    @validator("values")
    def check_values(cls, terms_values: List[str], values: Dict[str, Any]) -> List[str]:
        if len(set(terms_values)) != len(terms_values):
            name = values.get("name")
            raise ValueError(f"`TermsMetadataFilter` with name={name} cannot have repeated `values`")
        return terms_values

    @property
    def query_string(self) -> str:
        return f"{self.name}:{','.join(self.values)}"


class _NumericMetadataFilterSchema(MetadataFilterSchema):
    """Protected schema for the numeric `FeedbackDataset` metadata filters.

    Note:
        At least one of the `le` or `ge` attributes must be provided.

    Args:
        name: The name of the metadata property.
        le: The upper bound of the numeric value. Defaults to `None`.
        ge: The lower bound of the numeric value. Defaults to `None`.

    Disclaimer:
        You should not use this class directly, but instead use the classes that inherit
        from this one.
    """

    le: Optional[Union[int, float]] = None
    ge: Optional[Union[int, float]] = None

    _bounds_validator = root_validator(allow_reuse=True, skip_on_failure=True)(validate_numeric_metadata_filter_bounds)

    @property
    def query_string(self) -> str:
        filter_params = {}
        if self.le is not None:
            filter_params["le"] = self.le
        if self.ge is not None:
            filter_params["ge"] = self.ge
        return f"{self.name}:{filter_params}".replace("'", '"')


class IntegerMetadataFilter(_NumericMetadataFilterSchema):
    """Schema for the `FeedbackDataset` metadata filters of type `integer`. This kind
    of metadata filter will be used for filtering the metadata of a record based on
    an integer value to which `le` and `ge` filters can be applied.

    Note:
        At least one of the `le` or `ge` attributes must be provided.

    Args:
        name: The name of the metadata property.
        le: The upper bound of the integer value. Defaults to `None`.
        ge: The lower bound of the integer value. Defaults to `None`.

    Examples:
        >>> from argilla.client.feedback.schemas.metadata import IntegerMetadataFilter
        >>> IntegerMetadataFilter(name="day", le=15)
        >>> IntegerMetadataFilter(name="day", ge=15)
        >>> IntegerMetadataFilter(name="day", le=15, ge=10)
    """

    type: Literal[MetadataPropertyTypes.integer] = Field(MetadataPropertyTypes.integer.value, allow_mutation=False)
    le: Optional[int] = None
    ge: Optional[int] = None


class FloatMetadataFilter(_NumericMetadataFilterSchema):
    """Schema for the `FeedbackDataset` metadata filters of type `float`. This kind
    of metadata filter will be used for filtering the metadata of a record based on
    an float value to which `le` and `ge` filters can be applied.

    Note:
        At least one of the `le` or `ge` attributes must be provided.

    Args:
        name: The name of the metadata property.
        le: The upper bound of the float value. Defaults to `None`.
        ge: The lower bound of the float value. Defaults to `None`.

    Examples:
        >>> from argilla.client.feedback.schemas.metadata import FloatMetadataFilter
        >>> FloatMetadataFilter(name="price", le=15.0)
        >>> FloatMetadataFilter(name="price", ge=15.0)
        >>> FloatMetadataFilter(name="price", le=15.0, ge=10.0)
    """

    type: Literal[MetadataPropertyTypes.float] = Field(MetadataPropertyTypes.float.value, allow_mutation=False)
    le: Optional[float] = None
    ge: Optional[float] = None


MetadataFilters = Union[TermsMetadataFilter, IntegerMetadataFilter, FloatMetadataFilter]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import base64
import re
import warnings
from pathlib import Path
from typing import Callable, List, Optional, Union
from urllib.parse import urlparse

from argilla.utils.dependency import require_dependencies

# Define html supported types for each media
SUPPORTED_MEDIA_TYPES = {
    "video": ["mp4", "webm", "ogg"],
    "audio": ["mp3", "wav", "ogg"],
    "image": ["png", "jpg", "jpeg", "ico", "svg", "gif", "apng", "avif", "webp"],
}


def validate_media_type(media_type: str, file_type: str) -> None:
    """
    Utility function to validate if the given file type is supported by the given media type.

    Args:
        media_type: Type of media ('video', 'audio', or 'image').
        file_type: The type of the media file.
    """
    if file_type not in SUPPORTED_MEDIA_TYPES[media_type]:
        warnings.warn(
            f"This {file_type} might not be supported. Supported types for {media_type} are {SUPPORTED_MEDIA_TYPES[media_type]}",
            category=UserWarning,
        )

    if file_type == "ogg":
        warnings.warn("'ogg' files might not be supported in Safari.", category=UserWarning)


def get_file_data(
    file_source: Union[str, bytes], file_type: Optional[str] = None, media_type: Optional[str] = None
) -> bytes:
    """
    Utility function to check the input file and get the file data as bytes.

    Args:
        file_source: The path to the media file or a non-b64 encoded byte string.
        file_type: The type of the video file. If not provided, it will be inferred from the file extension.
        media_type: Type of media ('video', 'audio', or 'image').

    Returns:
        File data as bytes.

    Raises:
        FileNotFoundError: If the file does not exist or is empty.
        ValueError: If no provided file type using bytes as input, the file type does not match the expected extension or the file size is too large (>5MB).
    """
    if isinstance(file_source, bytes):
        if not file_type:
            raise ValueError("File type must be provided if file source is a byte string.")
        else:
            file_type != "pdf" and validate_media_type(media_type, file_type)
        file_data = file_source
    else:
        file_path = Path(file_source)
        if not file_path.exists() or file_path.stat().st_size == 0:
            raise FileNotFoundError(f"File {file_path} does not exist or is empty.")

        file_type = file_type or file_path.suffix[1:].lower()
        if file_path.suffix.lower() != f".{file_type}":
            raise ValueError(f"Provided file is not a {file_type.upper()}.")

        file_type != "pdf" and validate_media_type(media_type, file_type)

        file_data = file_path.read_bytes()

    if len(file_data) > 5_000_000:
        raise ValueError(
            f"File size is {len(file_data)} bytes. It is recommended to use files smaller than 5MB, as larger files might not render properly."
        )

    return file_data, file_type


def is_valid_dimension(dim: Optional[str]) -> bool:
    """
    Utility function to validate if the given dimension is a pixel or percentage value.

    Args:
        dim: The dimension string to validate (e.g., '300px', '50%').

    Returns:
        True if valid, False otherwise.
    """
    if dim is None:
        return True
    return bool(re.match(r"^\d+(px|%)$", dim))


def media_to_html(
    media_type: str,
    file_source: Union[str, bytes],
    file_type: Optional[str] = None,
    width: Optional[str] = None,
    height: Optional[str] = None,
    autoplay: Optional[bool] = False,
    loop: Optional[bool] = False,
) -> str:
    """
    Convert a media file to an HTML tag with embedded base64 data.

    Args:
        media_type: Type of media ('video', 'audio', or 'image').
        file_source: The path to the media file or a non-b64 encoded byte string.
        file_type: The type of the video file. If not provided, it will be inferred from the file extension.
        width: Display width in HTML. Defaults to None.
        height: Display height in HTML. Defaults to None.
        autoplay: True to autoplay media. Defaults to False.
        loop: True to loop media. Defaults to False.

    Returns:
        HTML tag with embedded base64 data.

    Raises:
        ValueError: If the width and height are not pixel or percentage.
    """
    if not is_valid_dimension(width) or not is_valid_dimension(height):
        raise ValueError("Width and height must be valid pixel (e.g., '300px') or percentage (e.g., '50%') values.")

    file_data, file_type = get_file_data(file_source, file_type, media_type)
    media_base64 = base64.b64encode(file_data).decode("utf-8")
    data_url = f"data:{media_type}/{file_type};base64,{media_base64}"

    common_attrs = f"{f' width={width}' if width else ''}{f' height={height}' if height else ''}"
    media_attrs = f"{' autoplay' if autoplay else ''}{' loop' if loop else ''}"

    if media_type == "video":
        return f"<video controls{common_attrs}{media_attrs}><source src='{data_url}' type='video/{file_type}'></video>"
    elif media_type == "audio":
        return f"<audio controls{media_attrs}><source src='{data_url}' type='audio/{file_type}'></audio>"
    elif media_type == "image":
        return f'<img src="{data_url}"{common_attrs}>'
    else:
        raise ValueError(f"Unsupported media type: {media_type}")


def video_to_html(
    file_source: Union[str, bytes],
    file_type: Optional[str] = None,
    width: Optional[str] = None,
    height: Optional[str] = None,
    autoplay: bool = False,
    loop: bool = False,
) -> str:
    """
    Convert a video file to an HTML tag with embedded base64 data.

    Args:
        file_source: The path to the media file or a non-b64 encoded byte string.
        file_type: The type of the video file. If not provided, it will be inferred from the file extension.
        width: Display width in HTML. Defaults to None.
        height: Display height in HTML. Defaults to None.
        autoplay: True to autoplay media. Defaults to False.
        loop: True to loop media. Defaults to False.

    Returns:
        The HTML tag with embedded base64 data.

    Examples:
        >>> from argilla.client.feedback.utils import video_to_html
        >>> html = video_to_html("my_video.mp4", width="300px", height="300px", autoplay=True, loop=True)
    """
    return media_to_html("video", file_source, file_type, width, height, autoplay, loop)


def audio_to_html(
    file_source: Union[str, bytes],
    file_type: Optional[str] = None,
    width: Optional[str] = None,
    height: Optional[str] = None,
    autoplay: bool = False,
    loop: bool = False,
) -> str:
    """
    Convert an audio file to an HTML tag with embedded base64 data.

    Args:
        file_source: The path to the media file or a non-b64 encoded byte string.
        file_type: The type of the audio file. If not provided, it will be inferred from the file extension.
        width: Display width in HTML. Defaults to None.
        height: Display height in HTML. Defaults to None.
        autoplay: True to autoplay media. Defaults to False.
        loop: True to loop media. Defaults to False.

    Returns:
        The HTML tag with embedded base64 data.

    Examples:
        >>> from argilla.client.feedback.utils import audio_to_html
        >>> html = audio_to_html("my_audio.mp3", width="300px", height="300px", autoplay=True, loop=True)
    """
    return media_to_html("audio", file_source, file_type, width, height, autoplay, loop)


def image_to_html(
    file_source: Union[str, bytes],
    file_type: Optional[str] = None,
    width: Optional[str] = None,
    height: Optional[str] = None,
) -> str:
    """
    Convert an image file to an HTML tag with embedded base64 data.

    Args:
        file_source: The path to the media file or a non-b64 encoded byte string.
        file_type: The type of the image file. If not provided, it will be inferred from the file extension.
        width: Display width in HTML. Defaults to None.
        height: Display height in HTML. Defaults to None.

    Returns:
        The HTML tag with embedded base64 data.

    Examples:
        >>> from argilla.client.feedback.utils import image_to_html
        >>> html = image_to_html("my_image.png", width="300px", height="300px")
    """
    return media_to_html("image", file_source, file_type, width, height)


def pdf_to_html(
    file_source: Union[str, bytes], width: Optional[str] = "1000px", height: Optional[str] = "1000px"
) -> str:
    """
    Convert a pdf file to an HTML tag with embedded data.

    Args:
        file_source: The path to the PDF file, a bytes object with PDF data, or a URL.
        width: Display width in HTML. Defaults to "1000px".
        height: Display height in HTML. Defaults to "1000px".

    Returns:
        HTML string embedding the PDF.

    Raises:
        ValueError: If the width and height are not pixel or percentage.

    Examples:
        >>> from argilla.client.feedback.utils import pdf_to_html
        >>> html = pdf_to_html("my_pdf.pdf", width="300px", height="300px")
    """
    if not is_valid_dimension(width) or not is_valid_dimension(height):
        raise ValueError("Width and height must be valid pixel (e.g., '300px') or percentage (e.g., '50%') values.")

    if isinstance(file_source, str) and urlparse(file_source).scheme in ["http", "https"]:
        return f'<embed src="{file_source}" type="application/pdf" width="{width}" height="{height}"></embed>'

    file_data, _ = get_file_data(file_source, "pdf")
    pdf_base64 = base64.b64encode(file_data).decode("utf-8")
    data_url = f"data:application/pdf;base64,{pdf_base64}"
    return f'<object id="pdf" data="{data_url}" type="application/pdf" width="{width}" height="{height}"><p>Unable to display PDF.</p></object>'


def create_token_highlights(
    tokens: List[str], weights: List[float], c_map: Optional[Union[str, Callable]] = "viridis"
) -> str:
    """
    Generates an HTML string with tokens highlighted based on the corresponding weights using a color map.

    Args:
        tokens: List of tokens to highlight.
        weights: List of weights corresponding to each token.
        c_map: Optional color map to use for highlighting. Can be a string (name of a matplotlib colormap)
                    or a callable that returns an RGB tuple

    Returns:
        An HTML string with the tokens highlighted.

    Raises:
        ValueError: If the length of tokens and weights is not the same.
        ValueError: If tokens or weights are empty.
        TypeError: If c_map is not a string or a callable.

    Examples:
        >>> from argilla.client.feedback.utils import create_token_highlights
        >>> tokens = ["This", "is", "a", "test"]
        >>> weights = [0.1, 0.2, 0.3, 0.4]
        >>> html = create_token_highlights(tokens, weights)
        >>> html = create_token_highlights(tokens, weights, c_map='viridis')
        >>> html = create_token_highlights(tokens, weights, c_map=custom_RGB)
    """
    require_dependencies("matplotlib")
    import matplotlib.pyplot as plt
    from matplotlib.colors import rgb2hex

    if len(tokens) != len(weights):
        raise ValueError("Length of tokens and weights must be the same.")

    if not tokens or not weights:
        raise ValueError("Token or weight lists must not be empty.")

    max_weight = max(weights)
    min_weight = min(weights)
    normalized_weights = [
        (w - min_weight) / (max_weight - min_weight) if max_weight != min_weight else 0 for w in weights
    ]

    html_str = ["<p>"]
    for token, weight in zip(tokens, normalized_weights):
        if weight == 0:
            html_str.append(f"{token} ")
            continue

        rgba = None

        if callable(c_map):
            rgba = c_map(weight)

        elif isinstance(c_map, str):
            cmap = plt.get_cmap(c_map)
            rgba = cmap(weight)[:3]

        else:
            raise TypeError("c_map must be either a string or a function.")

        hex_color = rgb2hex(rgba)
        html_str.append(f'<span style="background-color: {hex_color}">{token}</span> ')

    html_str.append("</p>")
    return "".join(html_str)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Dict, List, Optional, Union

from argilla.client.models import Text2TextRecord as ClientText2TextRecord
from argilla.client.sdk.commons.models import (
    MACHINE_NAME,
    BaseAnnotation,
    BaseRecord,
    PredictionStatus,
    ScoreRange,
    SdkVectorSearch,
    TaskStatus,
    UpdateDatasetRequest,
)
from argilla.pydantic_v1 import BaseModel, Field


class Text2TextPrediction(BaseModel):
    text: str
    score: float = Field(default=1.0, ge=0.0, le=1.0)


class Text2TextAnnotation(BaseAnnotation):
    sentences: List[Text2TextPrediction]


class CreationText2TextRecord(BaseRecord[Text2TextAnnotation]):
    text: str

    @classmethod
    def from_client(cls, record: ClientText2TextRecord):
        prediction = None
        if record.prediction is not None:
            prediction = Text2TextAnnotation(
                sentences=[
                    Text2TextPrediction(text=pred[0], score=pred[1])
                    if isinstance(pred, tuple)
                    else Text2TextPrediction(text=pred)
                    for pred in record.prediction
                ],
                agent=record.prediction_agent or MACHINE_NAME,
            )
        annotation = None
        if record.annotation is not None:
            annotation = Text2TextAnnotation(
                sentences=[Text2TextPrediction(text=record.annotation)],
                agent=record.annotation_agent or MACHINE_NAME,
            )

        return cls(
            text=record.text,
            prediction=prediction,
            annotation=annotation,
            vectors=cls._from_client_vectors(record.vectors),
            status=record.status,
            metadata=record.metadata,
            id=record.id,
            event_timestamp=record.event_timestamp,
        )


class Text2TextRecord(CreationText2TextRecord):
    last_updated: datetime = None
    _predicted: Optional[PredictionStatus] = Field(alias="predicted")

    def to_client(self) -> ClientText2TextRecord:
        return ClientText2TextRecord(
            text=self.text,
            prediction=[(sentence.text, sentence.score) for sentence in self.prediction.sentences]
            if self.prediction
            else None,
            prediction_agent=self.prediction.agent if self.prediction else None,
            annotation=self.annotation.sentences[0].text if self.annotation else None,
            annotation_agent=self.annotation.agent if self.annotation else None,
            vectors=self._to_client_vectors(self.vectors),
            status=self.status,
            metadata=self.metadata or {},
            id=self.id,
            event_timestamp=self.event_timestamp,
            metrics=self.metrics or None,
            search_keywords=self.search_keywords or None,
        )


class Text2TextBulkData(UpdateDatasetRequest):
    records: List[CreationText2TextRecord]


class Text2TextQuery(BaseModel):
    ids: Optional[List[Union[str, int]]]

    query_text: str = Field(default=None)

    annotated_by: List[str] = Field(default_factory=list)
    predicted_by: List[str] = Field(default_factory=list)

    vector: Optional[SdkVectorSearch] = Field(default=None)

    score: Optional[ScoreRange] = Field(default=None)

    status: List[TaskStatus] = Field(default_factory=list)

    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)
    metadata: Optional[Dict[str, Union[str, List[str]]]] = None
